from ...legacy.weave import registry_mem


def make_error_message(missing_ops, section_name):
    error_msg = f"{len(missing_ops)} Missing {section_name} Ops: \n"
    missing_ops.sort()
    no_prefix_ops = [op for op in missing_ops if "-" not in op]
    prefixed_ops = [op for op in missing_ops if "-" in op]
    error_msg += "* No Prefix:\n" + "\n".join(
        [f"  * [ ] `{op}`" for op in no_prefix_ops]
    )
    curr_prefix = None
    for op in prefixed_ops:
        op_prefix = op.split("-")[0]
        error_msg += "\n"
        if op_prefix != curr_prefix:
            curr_prefix = op_prefix
            error_msg += "* " + op_prefix + ":\n"
        error_msg += f"  * [ ] `{op}`"
    return error_msg


def test_op_coverage():
    ops = registry_mem.memory_registry.list_ops()
    all_py_names = set([op.name for op in ops])
    missing_visible_ops = list(
        visible_js_ops - js_op_exceptions - js_ops_with_new_names - all_py_names
    )
    missing_hidden_ops = list(
        hidden_js_ops - js_op_exceptions - js_ops_with_new_names - all_py_names
    )
    total_missing = len(missing_visible_ops) + len(missing_hidden_ops)
    error_msg = (
        f"**{total_missing} Missing Ops**\n\n"
        + make_error_message(missing_visible_ops, "Visible")
        + "\n\n"
        + make_error_message(missing_hidden_ops, "Hidden")
    )

    assert total_missing == 0, error_msg


js_ops_with_new_names = set(
    [
        "pick",
        "index",
        "dict-pick",
        "file-directUrlAsOf",
        "file-artifactVersion",
        "file-directUrl",
    ]
)

js_op_exceptions = set(
    [
        # Special Ops to support Weave1
        "type-__newType__",
        # We don't need to support converter panels (it is handled by the engine)
        "panel-projection",
        "panel-merge",
        "panel-runs-table",
        "panel-table",
        # Hidden ops with no product usage
        "if",
        "mapcount",
        "throwError",
        "get-tag",
        "internal-lambdaClosureArgBridge",
        "artifactMembership-id",
        "artifactType-sequences",
        "artifactType-portfolios",
        "artifactVersion-defaultFile",
        "artifactVersion-fileReturnType",
        "artifactVersion-state",
        "artifactVersion-updateAliasActions",
        "date_round-day",
        "date_round-hour",
        "date_round-minute",
        "date_round-quarter",
        "date_round-year",
        "date-sub",
        "dates-Equal",
        "dates-min",
        "entity-internalId",
        "Geo-selected",
        "Scatter-selected",
        "local-snowflaketable",
        "localArtifact-get",
        "number-roundtenth",
        "number-roundthousand",
        "number-roundHundredth",
        "op-non_none",
        "org-members",
        "project-reports",
        "project-run",  # This is used in the demo compute graph... surprised nowhere else
        "report-name",
        "report-description",
        "report-stargazers",
        "root-allEntities",
        "root-featuredreports",
        "root-localArtifact",
        "root_number-range",
        "tablerow-table",
        "type-string",
        "typedDict-values",
        "user-id",
        "user-runs",
        # Hidden ops only used in other op resolvers
        "list-PCA",
        "list-UMAP",
        "list-tSNE",
        "artifactVersion-historyStep",
        "root_number-binsfixed",
    ]
)


visible_js_ops = set(
    [
        "numbers-sum",
        "numbers-avg",
        "numbers-argmax",
        "numbers-argmin",
        "numbers-stddev",
        "numbers-min",
        "numbers-max",
        "number-add",
        "number-sub",
        "number-mult",
        "number-div",
        "number-modulo",
        "number-powBinary",
        "number-equal",
        "number-notEqual",
        "number-less",
        "number-greater",
        "number-lessEqual",
        "number-greaterEqual",
        "number-toString",
        "number-abs",
        "number-toTimestamp",
        "number-negate",
        "string-equal",
        "string-notEqual",
        "string-len",
        "string-add",
        "string-append",
        "string-prepend",
        "string-split",
        "string-partition",
        "string-startsWith",
        "string-endsWith",
        "string-isAlpha",
        "string-isNumeric",
        "string-isAlnum",
        "string-lower",
        "string-upper",
        "string-slice",
        "string-replace",
        "string-findAll",
        "string-contains",
        "string-strip",
        "string-lStrip",
        "string-rStrip",
        "string-levenshtein",
        "and",
        "or",
        "boolean-not",
        "boolean-not",
        "isNone",
        "pick",
        "entity-name",
        "entity-link",
        "file-size",
        "file-digest",
        "file-table",
        "file-joinedTable",
        "file-partitionedTable",
        "file-contents",
        "tag-project",
        "project-createdAt",
        "project-name",
        "project-runs",
        "project-artifactType",
        "project-artifactTypes",
        "project-artifact",
        "project-artifactVersion",
        "root-project",
        "tag-run",
        "run-name",
        "run-jobType",
        "run-user",
        "run-createdAt",
        "run-heartbeatAt",
        "run-config",
        "run-summary",
        "run-history",
        "run-loggedArtifactVersion",
        "run-loggedArtifactVersions",
        "run-usedArtifactVersions",
        "run-runtime",
        "artifactVersion-name",
        "artifactVersion-versionId",
        "artifactVersion-size",
        "artifactVersion-createdAt",
        "artifactVersion-files",
        "artifactVersion-file",
        "artifactVersion-usedBy",
        "artifactVersion-metadata",
        "artifactVersion-aliases",
        "artifactVersion-link",
        "asset-file",
        "count",
        "joinToStr",
        "index",
        "sort",
        "filter",
        "dropna",
        "map",
        "join",
        "concat",
        "contains",
        "artifact-name",
        "artifact-versions",
        "artifact-link",
        "artifactType-name",
        "artifactType-artifacts",
        "artifactType-artifactVersions",
        "table-rows",
        "partitionedtable-rows",
        "partitionedtable-rowsType",
        "partitionedtable-file",
        "joinedtable-rows",
        "joinedtable-rowsType",
        "joinedtable-file",
        "local-sqlconnection",
        "user-username",
        "panel-merge",
        "panel-table",
        "panel-projection",
        "panel-runs-table",
    ]
)

hidden_js_ops = set(
    [
        "list",
        "dict",
        "number-pow",
        "number-floor",
        "number-roundthousand",
        "number-roundHundredth",
        "number-roundtenth",
        "number-toFixed",
        "number-toByteString",
        "root_number-binsfixed",
        "numbers-binsequal",
        "number-bin",
        "root_number-range",
        "number-sin",
        "throwError",
        "string-in",
        "date-sub",
        "date-toNumber",
        "dates-min",
        "dates-equal",
        "date_round-year",
        "date_round-quarter",
        "date_round-month",
        "date_round-week",
        "date_round-day",
        "date_round-hour",
        "date_round-minute",
        "boolean-equal",
        "boolean-notEqual",
        "type-string",
        "dict-pick",
        "typedDict-values",
        "object-keytypes",
        "merge",
        "entity-internalId",
        "entity-isTeam",
        "entity-projects",
        "entity-portfolios",
        "entity-org",
        "file-type",
        "file-dir",
        "file-path",
        "file-directUrl",
        "file-directUrlAsOf",
        "file-media",
        "file-artifactVersion",
        "project-internalId",
        "project-entity",
        "project-link",
        "project-run",
        "project-filteredRuns",
        "project-reports",
        "project-artifacts",
        "project-runQueues",
        "root-entity",
        "root-artifactVersion",
        "root-org",
        "root-user",
        "root-viewer",
        "root-featuredreports",
        "root-allProjects",
        "root-allReports",
        "root-allArtifacts",
        "root-allEntities",
        "rpt_weekly_users_by_country_by_repo",
        "rpt_weekly_repo_users_by_persona",
        "rpt_weekly_engaged_user_count_by_repo",
        "rpt_repo_gpu_backends",
        "rpt_versus_other_repos",
        "rpt_runtime_buckets",
        "rpt_user_model_train_freq",
        "rpt_runs_versus_other_repos",
        "rpt_product_usage",
        "run-internalId",
        "run-id",
        "run-link",
        "run-project",
        "run-jobtype",
        "_run-historykeyinfo",
        "run-historyAsOf",
        "artifactVersion-state",
        "artifactVersion-digest",
        "artifactVersion-description",
        "artifactVersion-id",
        "artifactVersion-defaultFile",
        "artifactVersion-referenceCount",
        "artifactVersion-createdBy",
        "artifactVersion-createdByUser",
        "artifactVersion-hash",
        "artifactVersion-artifactSequence",
        "artifactVersion-artifactType",
        "artifactVersion-updateAliasActions",
        "artifactVersion-artifactCollections",
        "artifactVersion-memberships",
        "artifactVersion-historyStep",
        "artifactVersion-isWeaveObject",
        "artifactVersion-historyMetrics",
        "asset-artifactVersion",
        "mapcount",
        "offset",
        "limit",
        "list-createIndexCheckpointTag",
        "tag-indexCheckpoint",
        "mapEach",
        "groupby",
        "group-groupkey",
        "joinAll",
        "tag-joinObj",
        "unnest",
        "unique",
        "flatten",
        "list-tSNE",
        "list-UMAP",
        "list-PCA",
        "table-2DProjection",
        "range",
        "get-tag",
        "if",
        "internal-lambdaClosureArgBridge",
        "artifact-id",
        "artifact-description",
        "artifact-aliases",
        "artifact-type",
        "artifact-lastMembership",
        "artifact-createdAt",
        "artifact-project",
        "artifact-memberships",
        "artifact-membershipForAlias",
        "artifact-isPortfolio",
        "artifactAlias-alias",
        "artifactAlias-artifact",
        "artifactMembership-id",
        "artifactMembership-collection",
        "artifactMembership-artifactVersion",
        "artifactMembership-createdAt",
        "artifactMembership-commitHash",
        "artifactMembership-versionIndex",
        "artifactMembership-aliases",
        "artifactMembership-link",
        "artifactType-sequences",
        "artifactType-portfolios",
        "table-rowsType",
        "tablerow-table",
        "dir-pathReturnType",
        "dir-path",
        "dir-_as_w0_dict_",
        "localpathReturnType",
        "localpath",
        "root-localArtifact",
        "root-string",
        "localArtifact-get",
        "file-readcsv",
        "file-pandasreadcsv",
        "local-snowflaketable",
        "sqlconnection-tables",
        "sqlconnection-tablesType",
        "sqlconnection-table",
        "getReturnType",
        "get",
        "save",
        "artifactVersion-fileReturnType",
        "Ref-get",
        "execute",
        "function-__call__",
        "Scatter-selected",
        "Geo-selected",
        "Facet-selected",
        "Object-__getattr__",
        "type-__newType__",
        "op_get_tag_type",
        "op_make_type_tagged",
        "op_make_type_key_tag",
        "op-non_none",
        "none-coalesce",
        "org-members",
        "org-reports",
        "org-projects",
        "org-artifacts",
        "org-name",
        "org-teams",
        "report-name",
        "report-link",
        "report-description",
        "report-createdAt",
        "report-project",
        "report-creator",
        "report-viewcount",
        "report-stargazers",
        "runQueue-id",
        "user-id",
        "user-name",
        "user-email",
        "user-link",
        "user-runs",
        "user-entities",
        "normalizeUserCounts",
    ]
)



"""Basic autopatching of trackable libraries.

This module should not require any dependencies beyond the standard library. It should
check if libraries are installed and imported and patch in the case that they are.
"""


def autopatch() -> None:
    from ..integrations.anthropic.anthropic_sdk import anthropic_patcher
    from ..integrations.cerebras.cerebras_sdk import cerebras_patcher
    from ..integrations.cohere.cohere_sdk import cohere_patcher
    from ..integrations.dspy.dspy_sdk import dspy_patcher
    from ..integrations.groq.groq_sdk import groq_patcher
    from ..integrations.langchain.langchain import langchain_patcher
    from ..integrations.litellm.litellm import litellm_patcher
    from ..integrations.llamaindex.llamaindex import llamaindex_patcher
    from ..integrations.mistral import mistral_patcher
    from ..integrations.openai.openai_sdk import openai_patcher

    openai_patcher.attempt_patch()
    mistral_patcher.attempt_patch()
    litellm_patcher.attempt_patch()
    llamaindex_patcher.attempt_patch()
    langchain_patcher.attempt_patch()
    anthropic_patcher.attempt_patch()
    groq_patcher.attempt_patch()
    dspy_patcher.attempt_patch()
    cerebras_patcher.attempt_patch()
    cohere_patcher.attempt_patch()


def reset_autopatch() -> None:
    from ..integrations.anthropic.anthropic_sdk import anthropic_patcher
    from ..integrations.cerebras.cerebras_sdk import cerebras_patcher
    from ..integrations.cohere.cohere_sdk import cohere_patcher
    from ..integrations.dspy.dspy_sdk import dspy_patcher
    from ..integrations.groq.groq_sdk import groq_patcher
    from ..integrations.langchain.langchain import langchain_patcher
    from ..integrations.litellm.litellm import litellm_patcher
    from ..integrations.llamaindex.llamaindex import llamaindex_patcher
    from ..integrations.mistral import mistral_patcher
    from ..integrations.openai.openai_sdk import openai_patcher

    openai_patcher.undo_patch()
    mistral_patcher.undo_patch()
    litellm_patcher.undo_patch()
    llamaindex_patcher.undo_patch()
    langchain_patcher.undo_patch()
    anthropic_patcher.undo_patch()
    groq_patcher.undo_patch()
    dspy_patcher.undo_patch()
    cerebras_patcher.undo_patch()
    cohere_patcher.undo_patch()



# Synthetic data for model monitoring

import datetime
import os
import random
from datetime import timedelta

import numpy as np
import pandas as pd
import tqdm
from faker import Faker

from weave.legacy.weave.arrow.list_ import ArrowWeaveList, dataframe_to_arrow


# Generate the version schedule
def generate_version_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime
) -> dict:
    current_date = start_date
    versions = ["1.0"]
    version_schedule = {}
    while current_date <= end_date:
        date_versions = []
        for version in versions:
            service_percent = random.uniform(0, 1)
            date_versions.append((version, service_percent))

        version_schedule[current_date.date()] = date_versions
        current_date += timedelta(days=1)
        if random.random() < 0.10:  # 5% chance to introduce a new version each day
            new_version = f"{float(versions[-1])+0.1:.1f}"
            versions.append(new_version)
        if len(versions) > 1 and random.random() < 0.10:
            versions.pop(0)
    return version_schedule


# Generate the latency schedule
def generate_latency_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime
) -> dict:
    latency_schedule = {}
    for current_date in pd.date_range(start_date, end_date):
        base_latency = random.uniform(0.1, 1)
        day_factor = random.uniform(0.5, 1.5)
        month_factor = random.uniform(0.5, 1.5)
        latency = base_latency * day_factor * month_factor
        latency_schedule[current_date.date()] = latency
    return latency_schedule


# Generate the cost schedule
def generate_cost_schedule(
    start_date: datetime.datetime,
    end_date: datetime.datetime,
    cost_change_date: datetime.datetime,
) -> dict:
    cost_schedule = {}
    current_date = start_date
    cost_per_token = 0.01
    while current_date <= end_date:
        if current_date >= cost_change_date:
            cost_per_token = 0.005
        cost_schedule[current_date.date()] = cost_per_token
        current_date += timedelta(days=1)
    return cost_schedule


def generate_user_usage_schedule(
    start_date: datetime.datetime, end_date: datetime.datetime, users: list
) -> list:
    user_usage_schedule = []
    for user in users:
        current_date = start_date + timedelta(days=random.randrange(90))
        while current_date <= end_date:
            usage_periods = random.randint(1, 30)
            for _ in range(usage_periods):
                period_length_timedelta = timedelta(hours=random.randint(1, 24 * 7))
                rate = random.uniform(0.1, 10)
                user_usage_schedule.append(
                    (current_date, user, period_length_timedelta, rate)
                )
                current_date += period_length_timedelta  # Increment current_date
                if current_date > end_date:
                    break
    return user_usage_schedule


def random_predictions(n_users: int = 10) -> ArrowWeaveList:
    # Define our fake users
    fake = Faker()
    users = [fake.user_name() for _ in range(n_users)]

    # Read the file and generate prompts
    with open(
        os.path.abspath(
            os.path.join(os.path.dirname(__file__), "../..", "testdata/t8.shakespeare.txt")
        ),
        "r",
    ) as f:
        lines = f.read().split("\n")

    # Define the time range
    start_date = pd.to_datetime("2023-01-01", utc=True)
    end_date = pd.to_datetime("2023-03-31", utc=True)
    cost_change_date = pd.to_datetime("2023-02-15", utc=True)

    # Generate the schedules
    version_schedule = generate_version_schedule(start_date, end_date)
    latency_schedule = generate_latency_schedule(start_date, end_date)
    cost_schedule = generate_cost_schedule(start_date, end_date, cost_change_date)
    user_usage_schedule = generate_user_usage_schedule(start_date, end_date, users)

    # Helper function to generate a random completion
    def generate_completion(prompt: str) -> str:
        words = prompt.split()
        completion = " ".join(
            random.choices(words, k=int(len(words) * (random.random() + 0.1) * 10))
        )
        return completion

    data = []
    for usage in tqdm.tqdm(user_usage_schedule):
        usage_date, user, usage_period, rate = usage

        end_date = usage_date + usage_period
        increment = timedelta(hours=rate)

        while usage_date < end_date:
            # Find the version that was active during this usage
            if usage_date.date() not in version_schedule:
                break
            active_versions = version_schedule[usage_date.date()]
            # active_versions = [(version, percent) for date, version, percent in version_schedule if date.date() == usage_date.date()]
            # Normalize the service percentages
            total_percent = sum([percent for version, percent in active_versions])
            if total_percent == 0:
                continue
            normalized_percentages = [
                percent / total_percent for version, percent in active_versions
            ]

            version = np.random.choice(
                [v for v, p in active_versions], p=normalized_percentages
            )

            # Find the cost during this usage
            cost_per_token = cost_schedule[usage_date.date()]

            # Find the average latency during this usage
            latency = latency_schedule[usage_date.date()]
            latency *= 0.9 + random.random() * 0.2

            prompt = " ".join(random.sample(lines, 10))  # Increase prompt size
            completion = generate_completion(prompt)
            prompt_tokens = len(prompt.split())
            completion_tokens = len(completion.split())
            api_cost = (prompt_tokens + completion_tokens) * cost_per_token

            data.append(
                [
                    usage_date,
                    user,
                    version,
                    prompt,
                    completion,
                    prompt_tokens,
                    completion_tokens,
                    api_cost,
                    latency,
                ]
            )

            usage_date += increment

    df = pd.DataFrame(
        data,
        columns=[
            "timestamp",
            "username",
            "model_version",
            "prompt",
            "completion",
            "prompt_tokens",
            "completion_tokens",
            "api_cost",
            "latency",
        ],
    )
    return dataframe_to_arrow(df)  # type: ignore



"""Utility methods for converting exceptions to JSON."""

import json
import traceback
from typing import Optional, TypedDict


class StackFrameDict(TypedDict):
    filename: str
    line_number: Optional[int]
    function_name: str
    text: Optional[str]


class ExceptionDict(TypedDict, total=False):
    type: str
    message: str
    traceback: Optional[list[StackFrameDict]]


def frame_summary_to_dict(frame: traceback.FrameSummary) -> StackFrameDict:
    return {
        "filename": frame.filename,
        "line_number": frame.lineno,
        "function_name": frame.name,
        "text": frame.line,
    }


def exception_to_dict(exception: BaseException) -> ExceptionDict:
    result: ExceptionDict = {
        "type": type(exception).__name__,
        "message": str(exception),
    }
    if exception.__traceback__:
        stack = traceback.extract_stack(exception.__traceback__.tb_frame)[:-1]
        stack.extend(traceback.extract_tb(exception.__traceback__))
        result["traceback"] = [frame_summary_to_dict(fs) for fs in stack]
    return result


def exception_to_json_str(exception: BaseException) -> str:
    return json.dumps(exception_to_dict(exception))



import base64
import json
import os
import zlib

from weave.legacy.weave import serialize, storage
from weave.legacy.weave.server import handle_request


def test_graph_playback(dev_only_admin_env_override):
    for payload in execute_payloads:
        res = handle_request(payload, True, storage.make_js_serializer())
        res.results.unwrap()


def test_zlib_playback(dev_only_admin_env_override, use_server_gql_schema):
    if zlib_str == "":
        return
    req_bytes = zlib.decompress(base64.b64decode(zlib_str.encode("ascii")))
    json_data = json.loads(req_bytes)

    if filter_node != "":
        nodes = serialize.deserialize(json_data["graphs"])
        use_node = None
        for ndx, n in enumerate(nodes):
            if str(n) == filter_node:
                use_node = ndx
                break
        if use_node is None:
            raise Exception("Did not find filter node")
        json_data["graphs"]["targetNodes"] = [
            json_data["graphs"]["targetNodes"][use_node]
        ]

    execute_args = {
        "request": json_data,
        "deref": True,
        "serialize_fn": storage.make_js_serializer(),
    }

    response = handle_request(**execute_args)
    response.results.unwrap()


# (Only used in zlib test) - if you are testing a `WeaveNullishResponseException`, you can
# paste the stringified error node here and we will only execute that node.
filter_node = ""

zlib_str = ""
# Paste zlib into zlib.txt (from DD)
zlib_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "zlib.txt")
if os.path.exists(zlib_path):
    with open(zlib_path, "r") as f:
        zlib_str = f.read()
    prefix = "Execute request (zlib): "
    if zlib_str.startswith(prefix):
        zlib_str = zlib_str[len(prefix) :]

# Paste graphs into execute.json (from network console)
execute_payloads: list[dict] = []
execute_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "execute.json")
if os.path.exists(execute_path):
    with open(execute_path, "r") as f:
        execute_str = f.read()
    if execute_str and execute_str != "":
        execute_payloads: list[dict] = [json.loads(execute_str)]



import contextlib
import contextvars
import copy
import logging
import typing

if typing.TYPE_CHECKING:
    # from .run import Run
    from .weave_client import Call


class NoCurrentCallError(Exception): ...


_call_stack: contextvars.ContextVar[list["Call"]] = contextvars.ContextVar(
    "call", default=[]
)

logger = logging.getLogger(__name__)


def push_call(call: "Call") -> None:
    new_stack = copy.copy(_call_stack.get())
    new_stack.append(call)
    _call_stack.set(new_stack)


def pop_call(call_id: typing.Optional[str]) -> None:
    new_stack = copy.copy(_call_stack.get())
    if len(new_stack) == 0:
        logger.debug(
            f"weave pop_call error: Found empty callstack when popping call_id: {call_id}."
        )
        # raise ValueError("Call stack is empty")
        return
    if call_id:
        # assert that the call_id is in the stack
        for i in range(len(new_stack)):
            target_index = -(i + 1)
            call = new_stack[target_index]
            if call.id == call_id:
                # Actually do the slice
                # Note (Tim): I think this logic is not quite correct. This will
                # effectively pop off all calls up to and including the target
                # call. I think this is actually an error case. Throwing an
                # error here would disallow out-of-sequence call finishing, but
                # i think that might be a good thing.
                new_stack = new_stack[:target_index]
                break
        else:
            logger.debug(
                f"weave pop_call error: Call with id {call_id} not found in stack."
            )
            # raise ValueError(f"Call with id {call_id} not found in stack")
            return
    else:
        new_stack.pop()
    _call_stack.set(new_stack)


def require_current_call() -> "Call":
    """Get the Call object for the currently executing Op, within that Op.

    This allows you to access attributes of the Call such as its id or feedback
    while it is running.

    ```python
    @weave.op
    def hello(name: str) -> None:
        print(f"Hello {name}!")
        current_call = weave.require_current_call()
        print(current_call.id)
    ```

    It is also possible to access a Call after the Op has returned.

    If you have the Call's id, perhaps from the UI, you can use the `call` method on the
    `WeaveClient` returned from `weave.init` to retrieve the Call object.

    ```python
    client = weave.init("<project>")
    mycall = client.get_call("<call_id>")
    ```

    Alternately, after defining your Op you can use its `call` method. For example:

    ```python
    @weave.op
    def hello(name: str) -> None:
        print(f"Hello {name}!")

    mycall = hello.call("world")
    print(mycall.id)
    ```

    Returns:
        The Call object for the currently executing Op

    Raises:
        NoCurrentCallError: If tracking has not been initialized or this method is
            invoked outside an Op.
    """
    if (call := get_current_call()) is None:
        raise NoCurrentCallError(
            "Have you initialized weave and are you calling this from inside an op?"
        )
    return call


def get_current_call() -> typing.Optional["Call"]:
    """Get the Call object for the currently executing Op, within that Op.

    Returns:
        The Call object for the currently executing Op, or
        None if tracking has not been initialized or this method is
        invoked outside an Op.
    """
    return _call_stack.get()[-1] if _call_stack.get() else None


def get_call_stack() -> list["Call"]:
    return _call_stack.get()


@contextlib.contextmanager
def set_call_stack(
    stack: list["Call"],
) -> typing.Iterator[list["Call"]]:
    token = _call_stack.set(stack)
    try:
        yield stack
    finally:
        _call_stack.reset(token)



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import context_state as _context_state
from weave.legacy.weave import decorator_op
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def as OpDef


def make_tag_getter_op(
    tag_key: str,
    tag_type: types.Type = types.Any(),
    base_type: types.Type = types.Any(),
    op_name: typing.Optional[str] = None,
) -> "OpDef.OpDef":
    """Create an op that returns the value of a tag on a tagged value.

    Args:
        name: The name of the op.
        tag_key: The key of the tag to get.
        tag_type: The type of the tag to get.
        base_type: The type of the tagged value to get the tag from.

    Returns:
        The op.
    """
    from weave.legacy.weave.arrow.list_ import ArrowWeaveList, ArrowWeaveListType

    _loading_builtins_token = _context_state.set_loading_built_ins()

    # Uncomment below once we allow searching by type alone.
    # if tag_key is None and tag_type is None:
    #     raise ValueError("Must provide either tag_key or tag_type")
    if op_name is None:
        op_name = f"get_tag-{tag_key}"

    @decorator_op.op(  # type: ignore
        name=op_name,
        input_type={
            "obj": tagged_value_type.TaggedValueType(
                types.TypedDict({tag_key: types.optional(tag_type)}), base_type
            ),
        },
        output_type=lambda input_types: input_types["obj"].tag.property_types.get(
            tag_key, types.NoneType()
        ),
    )
    def tag_getter_op(obj):  # type: ignore
        return tag_store.find_tag(obj, tag_key, tag_type)

    # This is the vectorized version of the tag getter specifically for
    # ArrowWeaveList. We have discussed the possibility of having a single tag
    # getter op and a single vectorized tag getter op which can handle any tag
    # requested, but in the meantime, this matches the Weave0 pattern.
    @decorator_op.op(  # type: ignore
        name=f"ArrowWeaveList_{op_name}",
        input_type={
            "obj": ArrowWeaveListType(
                tagged_value_type.TaggedValueType(
                    types.TypedDict({tag_key: types.optional(tag_type)}), base_type
                ),
            )
        },
        output_type=lambda input_types: ArrowWeaveListType(
            input_types["obj"].object_type.tag.property_types[tag_key]
        ),
    )
    def awl_tag_getter_op(obj):  # type: ignore
        return ArrowWeaveList(
            obj._arrow_data.field("_tag").field(tag_key),
            obj.object_type.tag.property_types[tag_key],
            obj._artifact,
        )

    tag_getter_op._gets_tag_by_name = tag_key  # type: ignore
    awl_tag_getter_op._gets_tag_by_name = tag_key  # type: ignore

    _context_state.clear_loading_built_ins(_loading_builtins_token)

    return tag_getter_op  # type: ignore



import typing

import requests

import weave


@weave.type()
class SlurmJob:
    job_id: str
    comment: str
    job_state: str

    submit_time: int
    start_time: int
    end_time: int


@weave.op()
def jobs_render(
    jobs: weave.Node[list[SlurmJob]],
) -> weave.legacy.weave.panels.Table:
    return weave.legacy.weave.panels.Table(
        jobs,
        columns=[
            lambda job: job.job_id,
            lambda job: job.comment,
            lambda job: job.job_state,
            lambda job: job.submit_time,
            lambda job: job.start_time,
            lambda job: job.end_time,
            lambda job: job.end_time - job.start_time,
        ],
    )


@weave.type()
class SlurmNode:
    node_name: str
    state: str


@weave.op()
def nodes_render(
    nodes: weave.Node[list[SlurmNode]],
) -> weave.legacy.weave.panels.Table:
    return weave.legacy.weave.panels.Table(
        nodes,
        columns=[
            lambda node: node.node_name,
            lambda node: node.state,
        ],
    )


@weave.type()
class Slurm:
    restapi_url: str

    @property
    def full_url(self):
        return self.restapi_url + "/slurm/v0.0.37"

    @weave.op(pure=False)
    def jobs(self) -> list[SlurmJob]:
        resp = requests.get(self.full_url + "/jobs")
        jobs = resp.json()["jobs"]

        return [
            SlurmJob(
                job_id=j["job_id"],
                comment=j["comment"],
                job_state=j["job_state"],
                submit_time=j["submit_time"],
                start_time=j["start_time"],
                end_time=j["end_time"],
            )
            for j in reversed(jobs)
        ]

    @weave.op(pure=False)
    def nodes(self) -> list[SlurmNode]:
        resp = requests.get(self.full_url + "/nodes")
        nodes = resp.json()["nodes"]

        return [
            SlurmNode(node_name=n["name"], state=n["state"]) for n in reversed(nodes)
        ]


@weave.op(render_info={"type": "function"})
def slurm(restapi_url: str) -> Slurm:
    return Slurm(restapi_url)


@weave.op()
def slurm_render(
    slurm_node: weave.Node[Slurm],
) -> weave.legacy.weave.panels.Card:
    slurm = typing.cast(Slurm, slurm_node)
    return weave.legacy.weave.panels.Card(
        title="slurm",
        subtitle="",
        content=[
            weave.legacy.weave.panels.CardTab(
                name="Overview",
                content=weave.legacy.weave.panels.Group(
                    prefer_horizontal=True,
                    items=[
                        weave.legacy.weave.panels.LabeledItem(
                            item=slurm.jobs().count(), label="Total jobs"
                        ),
                        weave.legacy.weave.panels.LabeledItem(
                            item=slurm.nodes().count(), label="Total nodes"
                        ),
                    ],
                ),
            ),
            weave.legacy.weave.panels.CardTab(name="Nodes", content=slurm.nodes()),
            weave.legacy.weave.panels.CardTab(name="Jobs", content=slurm.jobs()),
        ],
    )



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave._dict_utils import typeddict_pick_output_type
from weave.legacy.weave.decorator_class import weave_class
from weave.legacy.weave.decorator_op import op
from weave.legacy.weave.language_features.tagging import tagged_value_type
from weave.legacy.weave.op_args import OpVarArgs
from weave.legacy.weave.ops_primitives._dict_utils import (
    tag_aware_dict_val_for_escaped_key,
)

# @op(
#     name='pick',
#     input_type={
#         'obj': types.TypedDict({}),
#         'key': types.String()},
#     output_type=types.Any())
# def pick(obj, key):
#     # TODO: is this how we want to delegate?
#     if hasattr(obj, 'pick'):
#         return obj.pick(key)
#     if isinstance(obj, list):
#         return [o.get(key) for o in obj]
#     return obj.get(key)

# TODO: make it so we can still call the underlying op pick!
#     or make it so we can declare pick on each item that has it?
#    The latter is much nicer if we can make it work I think.
#      should be easy, we can look up op based on which object its being
#      called on
# Then figure out how to do array/mapped ops...


def is_const_union_of_type(type_, of_type):
    if not isinstance(type_, types.UnionType):
        return False
    return all(
        isinstance(m, types.Const) and m.val_type == of_type for m in type_.members
    )


# TODO: in this PR, this was important for making scenario compare work.
#    - but Tim did some similar fix for Union Types, so I need to revisit
# def typeddict_pick_output_type(input_types):
#     if isinstance(input_types["self"], types.UnionType):
#         return types.union(
#             *(
#                 typeddict_pick_output_type({"self": m, "key": input_types["key"]})
#                 for m in input_types["self"].members
#             )
#         )
#     property_types = input_types["self"].property_types
#     if is_const_union_of_type(input_types["key"], types.String()):
#         member_types = []
#         for m in input_types["key"].members:
#             prop_type = property_types.get(m.val)
#             if prop_type is None:
#                 member_types.append(types.NoneType())
#             else:
#                 member_types.append(prop_type)
#         return types.union(*member_types)

#     if not isinstance(input_types["key"], types.Const):
#         return types.union(*list(property_types.values()))
#     key = input_types["key"].val
#     output_type = property_types.get(key)
#     if output_type is None:
#         # TODO: we hack this to types.Number() for now! This is relied
#         # on by tests because readcsv() doesn't properly return a full
#         # type right now. Super janky
#         return types.Number()
#     return output_type


@weave_class(weave_type=types.TypedDict)
class TypedDict:
    def __setitem__(self, k, v, action=None):
        dict.__setitem__(self, k, v)
        return self

    @op(
        setter=__setitem__,
        input_type={"self": types.TypedDict({}), "key": types.optional(types.String())},
        output_type=typeddict_pick_output_type,
    )
    def pick(self, key):
        return tag_aware_dict_val_for_escaped_key(self, key)

    @op(
        output_type=lambda input_type: types.List(
            types.union(
                *(
                    types.Const(types.String(), k)
                    for k in input_type["self"].property_types.keys()
                )
            )
        )
    )
    def keys(self):
        return list(self.keys())

    @op(
        name="merge",
        input_type={"lhs": types.TypedDict({}), "rhs": types.TypedDict({})},
        output_type=lambda input_types: types.TypedDict(
            {**input_types["lhs"].property_types, **input_types["rhs"].property_types}
        ),
    )
    def merge(lhs, rhs):
        return {**lhs, **rhs}

    @op(
        name="typedDict-values",
        output_type=lambda input_types: types.List(
            types.union(*input_types["self"].property_types.values())
        ),
    )
    def values(self):
        return self.values()

    __getitem__ = pick


def dict_return_type(input_types):
    # Discard Const types!
    # TODO: this is probably not really ideal. We lose a lot of information here.
    # But we don't handle unions very well in Weave1 right now, so this makes
    # developing panel composition stuff easier.
    res = {}
    for k, v in input_types.items():
        if isinstance(v, types.Const):
            res[k] = v.val_type
        else:
            res[k] = v
    return types.TypedDict(res)


@op(
    name="dict",
    input_type=OpVarArgs(types.Any()),
    # output_type=lambda input_types: types.TypedDict(input_types),
    output_type=dict_return_type,
)
def dict_(**d):
    return d


def _keytypes_resovler(
    weave_type: types.Type,
) -> list[dict[str, typing.Union[str, types.Type]]]:
    res: list[dict[str, typing.Union[str, types.Type]]] = []
    if isinstance(weave_type, types.TypedDict):
        for k, v in weave_type.property_types.items():
            res.append(
                {
                    "key": k,
                    "type": v,
                }
            )
    elif isinstance(weave_type, tagged_value_type.TaggedValueType):
        return _keytypes_resovler(weave_type.value)
    return res


@op(
    name="object-keytypes",
    input_type={
        "obj": types.TypedDict({}),
    },
    output_type=types.List(
        types.TypedDict(
            {
                "key": types.String(),
                "type": types.TypeType(),
            }
        )
    ),
)
def object_keytypes(obj):
    # Unlike Weave0, we don't have the type information of inputs, so
    # unfortunately we have to figure out the types using the data....
    weave_type = types.TypeRegistry.type_of(obj)
    return _keytypes_resovler(weave_type)



import pytest

import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave import graph, weave_internal
from weave.legacy.weave.dispatch import _dispatch_first_arg, _resolve_op_ambiguity
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_domain import wb_domain_types

_loading_builtins_token = _context.set_loading_built_ins()


@weave.op()
def add_1(x: int) -> int:
    return x + 1


@weave.op(name="string-special_mult")
def mult_string(x: int, y: str) -> str:
    return y * x


@weave.op(name="number-special_mult")
def mult_number(x: int, y: int) -> int:
    return y * x


_context.clear_loading_built_ins(_loading_builtins_token)


def test_chain():
    a = weave.save(1)
    res = add_1(a)
    assert weave.use(res) == 2
    res = a.add_1()
    assert weave.use(res) == 2


def test_dispatch_params():
    x = weave.save(2)
    num = weave.save(3)
    string = weave.save("a")

    assert weave.use(x.special_mult(string)) == "aa"
    assert weave.use(x.special_mult(num)) == 6


# def test_model_type():
#     dummy_model = weave_internal.make_const_node(
#         weave_keras.KerasModel.make_type(
#             [weave_keras.KerasTensorType.from_list([None, 1], weave.types.String())],
#             [weave_keras.KerasTensorType.from_list([None, 1], weave.types.Number())],
#         ),
#         None,
#     )
#     assert dummy_model.call_string("a").type == weave.types.List(weave.types.Number())
#     dummy_model = weave_internal.make_const_node(
#         weave_keras.KerasModel.make_type(
#             [weave_keras.KerasTensorType.from_list([None, 1], weave.types.String())],
#             [weave_keras.KerasTensorType.from_list([None, 1], weave.types.String())],
#         ),
#         None,
#     )
#     assert dummy_model.call_string("a").type == weave.types.List(weave.types.String())


def test_pick_map():
    a = weave.save([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    res = a.pick("a")
    assert weave.use(res) == [1, 3]


def test_json_pick_map():
    res = weave.legacy.weave.graph.OutputNode.from_json(
        {
            "nodeType": "output",
            "type": {"type": "list", "objectType": "number"},
            "fromOp": {
                "name": "pick",
                "inputs": {
                    "obj": {
                        "nodeType": "output",
                        "type": {
                            "type": "list",
                            "objectType": {
                                "type": "typedDict",
                                "propertyTypes": {"a": "number", "b": "number"},
                            },
                        },
                        "fromOp": {
                            "name": "list",
                            "inputs": {
                                "0": {
                                    "nodeType": "output",
                                    "type": {
                                        "type": "typedDict",
                                        "propertyTypes": {"a": "number", "b": "number"},
                                    },
                                    "fromOp": {
                                        "name": "dict",
                                        "inputs": {
                                            "a": {
                                                "nodeType": "const",
                                                "type": "number",
                                                "val": 1,
                                            },
                                            "b": {
                                                "nodeType": "const",
                                                "type": "number",
                                                "val": 2,
                                            },
                                        },
                                    },
                                }
                            },
                        },
                    },
                    "key": {"nodeType": "const", "type": "string", "val": "a"},
                },
            },
        }
    )

    assert weave.use(res) == [1]


def test_nested_js_dict_pick():
    assert (
        weave.use(
            graph.OutputNode(
                weave.types.Number(),
                "pick",
                {
                    "obj": graph.OutputNode(
                        weave.types.TypedDict(
                            {
                                "e": weave.types.Number(),
                                "f": weave.types.Number(),
                                "j": weave.types.Number(),
                            }
                        ),
                        "pick",
                        {
                            "obj": graph.ConstNode(
                                weave.types.TypedDict(
                                    {
                                        "a": weave.types.Number(),
                                        "b": weave.types.Number(),
                                        "c": weave.types.TypedDict(
                                            {
                                                "e": weave.types.Number(),
                                                "f": weave.types.Number(),
                                                "j": weave.types.Number(),
                                            }
                                        ),
                                    }
                                ),
                                {"a": 1, "b": 2, "c": {"d": 3, "e": 4, "f": 5}},
                            ),
                            "key": graph.ConstNode(weave.types.String(), "c"),
                        },
                    ),
                    "key": graph.ConstNode(weave.types.String(), "f"),
                },
            )
        )
        == 5
    )


def test_dispatch_of_ambiguous_ops():
    artifacts_node = weave.legacy.weave.ops.project("a", "b").artifacts()
    # just making this call is the test since it is
    # not clear if it should be artifact-project or
    # tag-project
    project_node = artifacts_node[0].project()
    projects_node = artifacts_node.project()


@pytest.mark.parametrize(
    "name, first_arg_type, exp_op_name",
    [
        ("contains", weave.types.String(), "string-contains"),
        ("contains", weave.types.List(weave.types.String()), "contains"),
        (
            "project",
            TaggedValueType(
                weave.types.TypedDict({"project": wb_domain_types.ProjectType}),
                wb_domain_types.RunType,
            ),
            "run-project",
        ),
    ],
)
def test_get_op_for_inputs(name, first_arg_type, exp_op_name):
    assert (
        _resolve_op_ambiguity(
            _dispatch_first_arg(name, first_arg_type), first_arg_type
        ).name
        == exp_op_name
    )



import numpy as np
from tensorflow import keras
from tensorflow.keras import layers
from PIL import Image
import numpy as np

# Model / data parameters
num_classes = 10
input_shape = (28, 28, 1)


def image_from_array(image_arr):
    return Image.fromarray((image_arr.reshape(28, 28) * 255).astype("uint8")).resize(
        (112, 112)
    )


def get_dataset():
    # Load the data and split it between train and test sets
    (x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()

    # Scale images to the [0, 1] range
    x_train = x_train.astype("float32") / 255
    x_test = x_test.astype("float32") / 255
    # Make sure images have shape (28, 28, 1)
    x_train = np.expand_dims(x_train, -1)
    x_test = np.expand_dims(x_test, -1)
    print("x_train shape:", x_train.shape)
    print(x_train.shape[0], "train samples")
    print(x_test.shape[0], "test samples")

    return x_train, y_train, x_test, y_test


def train_model(
    x_train, y_train, x_test, y_test, conv_layers=2, batch_size=128, epochs=15
):
    # Exmaple from: https://keras.io/examples/vision/mnist_convnet/

    # convert class vectors to binary class matrices
    y_train = keras.utils.to_categorical(y_train, num_classes)
    y_test = keras.utils.to_categorical(y_test, num_classes)

    s_layers = [keras.Input(shape=input_shape)]
    for i in range(conv_layers):
        s_layers.append(layers.Conv2D(32, kernel_size=(3, 3), activation="relu"))
        s_layers.append(layers.MaxPooling2D(pool_size=(2, 2)))
    s_layers.append(layers.Flatten())
    s_layers.append(layers.Dropout(0.5))
    s_layers.append(layers.Dense(num_classes, activation="softmax"))
    model = keras.Sequential(s_layers)

    print(model.summary())

    model.compile(
        loss="categorical_crossentropy", optimizer="adam", metrics=["accuracy"]
    )

    model.fit(
        x_train, y_train, batch_size=batch_size, epochs=epochs, validation_split=0.1
    )

    score = model.evaluate(x_test, y_test, verbose=0)
    print("Test loss:", score[0])
    print("Test accuracy:", score[1])

    return model


import gradio as gr


def render_interface(predict):
    def on_predict(pil_image):
        if pil_image is None:
            return None
        return predict(pil_image)

    def on_correct(pil_image, prediction):
        pass

    def on_incorrect(pil_image, prediction):
        pass

    with gr.Blocks() as demo:
        with gr.Row():
            inp = gr.Sketchpad(
                label="Draw a number between 0 and 9",
                brush_radius=5,
                type="pil",
                shape=(128, 128),
            )
            out = gr.Label(label="Prediction", num_top_classes=4)
        with gr.Row():
            with gr.Column():
                btn = gr.Button("Predict")
            with gr.Row():
                btn2 = gr.Button("Correct   ✅", variant="primary")
                btn3 = gr.Button("Incorrect   🚫", variant="stop")
        btn.click(fn=on_predict, inputs=inp, outputs=out)
        btn2.click(fn=on_correct, inputs=[inp, out], outputs=[])
        btn3.click(fn=on_incorrect, inputs=[inp, out], outputs=[])

    demo.launch(quiet=True, show_api=False, height=320)



import dataclasses

import pytest

import weave
from weave.legacy.weave import artifact_local, storage
from weave.legacy.weave.ops_arrow import to_arrow
from weave.legacy.weave.ops_domain import wbmedia

# This is not a valid artifact, but we need one to test. We set _read_dirname
# so that the artifact's is_saved property is True, so that everything works
# as expected.
UNSAVED_TEST_ARTIFACT = artifact_local.LocalArtifact("artifact2", "latest")
UNSAVED_TEST_ARTIFACT._read_dirname = "x"

CONCAT_TEST_CASES = [
    ([1, 2, 3], [4, 5, 6]),
    ([1, 2, 3], ["a", "b", "c"]),
    ([1, 2, 3], ["a", "b", "c", False]),
    ([1, 2, "a", {"b": 7}, {"d": 8}], [4, False, None, {"d": 8}, {"c": 9}]),
    (
        [1, None, 2, "a", {"b": 7}, {"d": 8}, {"d": None}, {"d": "j", "b": "s"}],
        [4, False, None, {"d": 8}, {"c": 9}],
    ),
    ([None, 1], ["a", None]),
    ([{}, None], [{}, None, 9]),
    ([{"a": {}}], [{"a": {}}]),
    ([1, 2, 3], [{"a": 5}]),
    ([{"a": 5}], [{"a": 9, "b": 6}]),
    ([1, 2, 3], [[4, 5], [6, 7, 8]]),
    ([[1, 2, 3]], [[4, 5], [6, 7, 8]]),
    ([{"a": [0.5]}, {"a": 8}], [{}]),
    ([{"a": [0.5]}, {"a": 8}], [{}, {"a": ["a", "b", None]}]),
    ([["a", "b"], ["c"]], [[4, 5], [6, 7, 8]]),
    ([["a", "b"], [{"x": 15}]], [[4, 5], [6, 7, 8]]),
    (
        [1, 2, 3],
        [
            wbmedia.ImageArtifactFileRef(
                UNSAVED_TEST_ARTIFACT,
                "path",
                "format",
                25,
                35,
                "sha256",
                None,
                None,
                None,
            )
        ],
    ),
    (
        [
            wbmedia.ImageArtifactFileRef(
                UNSAVED_TEST_ARTIFACT,
                "path1",
                "format1",
                25,
                35,
                "sha256-1",
                {
                    "box": [
                        {
                            "position": {
                                "maxX": 5.0,
                                "maxY": 9.0,
                                "minX": 10.0,
                                "minY": 15.0,
                            },
                            "class_id": 0,
                            "scores": {"a": 0.5},
                        }
                    ]
                },
                None,
                None,
            )
        ],
        [
            wbmedia.ImageArtifactFileRef(
                UNSAVED_TEST_ARTIFACT,
                "path2",
                "format2",
                50,
                70,
                "sha256-2",
                None,
                None,
                None,
            )
        ],
    ),
    (
        [{"a": 5, "b": None}],
        [
            {
                "a": 7,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    50,
                    70,
                    "sha256-2",
                    None,
                    None,
                    None,
                ),
            }
        ],
    ),
    (
        [{"a": 5, "b": None}],
        [
            {
                "a": 7,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    50,
                    70,
                    "sha256-2",
                    {
                        "box": [
                            {
                                "position": {
                                    "maxX": 5.0,
                                    "maxY": 9.0,
                                    "minX": 10.0,
                                    "minY": 15.0,
                                },
                                "class_id": 0,
                                "scores": {"a": 0.5},
                            }
                        ]
                    },
                    None,
                    None,
                ),
            }
        ],
    ),
    (
        [{"a": 5, "b": None}],
        [
            {
                "a": 49,
                "b": None,
            },
            {
                "a": 7,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    50,
                    70,
                    "sha256-2",
                    {
                        "box": [
                            {
                                "position": {
                                    "maxX": 5.0,
                                    "maxY": 9.0,
                                    "minX": 10.0,
                                    "minY": 15.0,
                                },
                                "class_id": 0,
                                "scores": {"a": 0.5},
                            }
                        ]
                    },
                    None,
                    None,
                ),
            },
        ],
    ),
    (
        [
            {
                "a": 7,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    50,
                    70,
                    "sha256-2",
                    {
                        "box": [
                            {
                                "position": {
                                    "maxX": 5.0,
                                    "maxY": 9.0,
                                    "minX": 10.0,
                                    "minY": 15.0,
                                },
                                "class_id": 0,
                                "scores": {"a": 0.5},
                            }
                        ]
                    },
                    None,
                    None,
                ),
            },
            {
                "a": 45,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    50,
                    70,
                    "sha256-25",
                    {"box": []},
                    None,
                    None,
                ),
            },
        ],
        [
            {"a": 10, "b": None},
            {
                "a": 11,
                "b": wbmedia.ImageArtifactFileRef(
                    UNSAVED_TEST_ARTIFACT,
                    "path2",
                    "format2",
                    67,
                    14,
                    "sha256-3",
                    {"box": []},
                    None,
                    None,
                ),
            },
        ],
    ),
]


def fix_for_compare(x1):
    if isinstance(x1, list):
        return [fix_for_compare(x) for x in x1]
    elif isinstance(x1, dict):
        return {k: fix_for_compare(v) for k, v in x1.items()}
    elif dataclasses.is_dataclass(x1):
        return x1.__class__(
            **{k: fix_for_compare(v) for k, v in dataclasses.asdict(x1).items()}
        )
    elif isinstance(x1, artifact_local.LocalArtifact):
        # Same hack as what we did to construct UNSAVED_TEST_ARTIFACT above.
        if x1._read_dirname is None:
            x1._read_dirname = "x"
        return (x1.name, x1.version)
    return x1


@pytest.mark.parametrize(
    "l1, l2",
    CONCAT_TEST_CASES,  # ids=lambda arg: str(weave.type_of(arg).object_type)
)
def test_extend(l1, l2):
    print()
    print("L1", l1)
    print("L2", l2)
    a1 = to_arrow(l1)
    a1_ref = storage.save(a1)
    a1_again = storage.get(str(a1_ref))
    assert a1.object_type == a1_again.object_type
    assert a1._arrow_data == a1_again._arrow_data
    a1 = a1_again

    a2 = to_arrow(l2)
    a2_ref = storage.save(a2)
    a2_again = storage.get(str(a2_ref))
    assert a2.object_type == a2_again.object_type
    assert a2._arrow_data == a2_again._arrow_data
    a2 = a2_again

    print("L1.object_type", a1.object_type)
    print("L2.object_type", a2.object_type)
    result = a1.concat(a2)
    expected = to_arrow(l1 + l2)
    print("EXPECTED TYPE", expected.object_type)
    print("RESULT TYPE", result.object_type)

    print("EXPECTED", expected.to_pylist_tagged())
    print("RESULT", result.to_pylist_tagged())
    # print("RESULT ARROW", result._arrow_data)
    # print("EXP ARROW", expected._arrow_data)

    # assert result.object_type == expected.object_type
    assert fix_for_compare(result.to_pylist_tagged()) == fix_for_compare(
        expected.to_pylist_tagged()
    )

    # Not working yet because of null handling
    # assert result._arrow_data == expected._arrow_data
    # assert_concat_result_equal(result.to_pylist_notags(), expected.to_pylist_notags())


CONCAT_WBTYPED_TEST_CASES = [
    # Number/Int/Floats
    # Int / Int Data, Permutations of Weave Types
    ([1, 2, 3], [4, 5, 6], weave.types.Number(), weave.types.Number()),
    ([1, 2, 3], [4, 5, 6], weave.types.Number(), weave.types.Int()),
    ([1, 2, 3], [4, 5, 6], weave.types.Int(), weave.types.Number()),
    ([1, 2, 3], [4, 5, 6], weave.types.Int(), weave.types.Int()),
    # Int / Float Data, Permutations of Weave Types
    ([1, 2, 3], [4.5, 5.5, 6.5], weave.types.Number(), weave.types.Number()),
    ([1, 2, 3], [4.5, 5.5, 6.5], weave.types.Number(), weave.types.Float()),
    ([1, 2, 3], [4.5, 5.5, 6.5], weave.types.Int(), weave.types.Number()),
    ([1, 2, 3], [4.5, 5.5, 6.5], weave.types.Int(), weave.types.Float()),
    # Float / Int Data, Permutations of Weave Types
    ([1.5, 2.5, 3.5], [4, 5, 6], weave.types.Number(), weave.types.Number()),
    ([1.5, 2.5, 3.5], [4, 5, 6], weave.types.Number(), weave.types.Int()),
    ([1.5, 2.5, 3.5], [4, 5, 6], weave.types.Float(), weave.types.Number()),
    ([1.5, 2.5, 3.5], [4, 5, 6], weave.types.Float(), weave.types.Int()),
    # Float / Float Data, Permutations of Weave Types
    ([1.5, 2.5, 3.5], [4.5, 5.5, 6.5], weave.types.Number(), weave.types.Number()),
    ([1.5, 2.5, 3.5], [4.5, 5.5, 6.5], weave.types.Number(), weave.types.Float()),
    ([1.5, 2.5, 3.5], [4.5, 5.5, 6.5], weave.types.Float(), weave.types.Number()),
    ([1.5, 2.5, 3.5], [4.5, 5.5, 6.5], weave.types.Float(), weave.types.Float()),
]


@pytest.mark.parametrize(
    "l1, l2, l1_wb_type, l2_wb_type",
    CONCAT_WBTYPED_TEST_CASES,
)
def test_concat_wbtyped(l1, l2, l1_wb_type, l2_wb_type):
    print()
    print("L1", l1)
    print("L2", l2)

    if l1_wb_type:
        l1_wb_type = weave.types.List(l1_wb_type)
    if l2_wb_type:
        l2_wb_type = weave.types.List(l2_wb_type)

    a1 = to_arrow(l1, l1_wb_type)
    a1_ref = storage.save(a1)
    a1_again = storage.get(str(a1_ref))
    assert a1.object_type == a1_again.object_type
    assert a1._arrow_data == a1_again._arrow_data
    a1 = a1_again

    a2 = to_arrow(l2, l2_wb_type)
    a2_ref = storage.save(a2)
    a2_again = storage.get(str(a2_ref))
    assert a2.object_type == a2_again.object_type


def test_image_ref():
    py = wbmedia.ImageArtifactFileRef(
        "artifact1",
        "path1",
        "format1",
        25,
        35,
        "sha256-1",
        None,
        None,
        None,
    )
    a = to_arrow([py])
    py2 = a.to_pylist_tagged()[0]
    assert py == py2


import random


def make_random_python_scalar():
    # Generate random python scalar
    rand = random.random()
    if rand < 0.2:
        return None
    elif rand < 0.4:
        return random.randint(0, 100)
    elif rand < 0.6:
        return random.random()
    elif rand < 0.8:
        return random.choice(["a", "b", "c"])
    else:
        return random.choice([True, False])


def make_random_python_list(depth=0):
    # Generate random python list
    return [
        make_random_python_val(depth=depth + 1) for _ in range(random.randint(0, 3))
    ]


def make_random_python_dict(depth=0):
    # Generate random python dict
    return {
        str(i): make_random_python_val(depth=depth + 1)
        for i in range(random.randint(0, 3))
    }


def make_random_python_val(depth=0):
    # Generate random python data with a mix of types
    rand = random.random()
    if depth > 3 or rand < 0.4:
        return make_random_python_scalar()
    elif rand < 0.7:
        return make_random_python_list(depth=depth + 1)
    else:
        return make_random_python_dict(depth=depth + 1)


random.seed(0)
random_cases = []
for i in range(100):
    l1 = [make_random_python_val() for i in range(3)]
    l2 = [make_random_python_val() for i in range(3)]
    random_cases.append((l1, l2))


@pytest.mark.parametrize("l1, l2", random_cases)
def test_random(l1, l2):
    print("L1", l1)
    print("L2", l2)
    a1 = to_arrow(l1)
    a2 = to_arrow(l2)
    result = a1.concat(a2)
    expected = to_arrow(l1 + l2)
    print("L1 OBJECT TYPE", a1.object_type)
    print("L2 OBJECT TYPE", a2.object_type)

    print("EXPECTED OBJECT TYPE", expected.object_type)
    print("RESULT OBJECT TYPE", result.object_type)
    # assert result.object_type.asdict() == expected.object_type.asdict()

    print("EXPECTED PY", expected.to_pylist_notags())
    print("RESULT PY", result.to_pylist_notags())

    assert fix_for_compare(result.to_pylist_tagged()) == fix_for_compare(
        expected.to_pylist_tagged()
    )


CONVERT_TEST_CASES = [
    ([{}, {"a": 5}], [{"a": None}, {"a": 5}]),
    ([{}, {"a": 5}, 92], [{"a": None}, {"a": 5}, 92]),
]


@pytest.mark.parametrize("l,expected", CONVERT_TEST_CASES)
def test_convert(l, expected):
    a = to_arrow(l)
    l2 = a.to_pylist_tagged()
    assert l2 == expected



import weave
from weave.legacy.weave import panel


@weave.type("traceSpanPanel")
class TraceSpanPanel(panel.Panel):
    id = "traceSpanPanel"


@weave.type("traceSpanModelPanel")
class TraceSpanModelPanel(panel.Panel):
    id = "traceSpanModelPanel"



import dataclasses
import json
import math

import numpy
import numpy as np
import pandas
import pyarrow as pa
import pyarrow.parquet as pq

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class
from weave.legacy.weave import box, file_base, graph, mappers_python
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type
from weave.legacy.weave.ops_primitives import list_


# Hack hack hack
# TODO: we can do this with Weave instead of writing a giant switch statement
#   (opNumberGreater can operate on Dataframe)
def filter_fn_to_pandas_filter(df, filter_fn_node):
    if isinstance(filter_fn_node, graph.ConstNode):
        return filter_fn_node.val
    elif isinstance(filter_fn_node, graph.OutputNode):
        op_name = graph.op_full_name(filter_fn_node.from_op)
        if op_name == "number-greater":
            return filter_fn_to_pandas_filter(
                df, filter_fn_node.from_op.inputs["lhs"]
            ) > filter_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["rhs"])
        elif op_name == "pick":
            return filter_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["obj"])[
                filter_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["key"])
            ]
        elif op_name == "typedDict-pick":
            return filter_fn_to_pandas_filter(
                df, filter_fn_node.from_op.inputs["self"]
            )[filter_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["key"])]
        raise Exception("unhandled op name", op_name)
    elif isinstance(filter_fn_node, graph.VarNode):
        if filter_fn_node.name == "row":
            return df
        raise Exception("unhandled var name")


def groupby_fn_to_pandas_filter(df, filter_fn_node):
    if isinstance(filter_fn_node, graph.ConstNode):
        return filter_fn_node.val
    elif isinstance(filter_fn_node, graph.OutputNode):
        op_name = graph.op_full_name(filter_fn_node.from_op)
        if op_name == "number-greater":
            return groupby_fn_to_pandas_filter(
                df, filter_fn_node.from_op.inputs["lhs"]
            ) > groupby_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["rhs"])
        elif op_name == "pick":
            return groupby_fn_to_pandas_filter(
                df, filter_fn_node.from_op.inputs["obj"]
            )[groupby_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["key"])]
        elif op_name == "typedDict-pick":
            return groupby_fn_to_pandas_filter(
                df, filter_fn_node.from_op.inputs["self"]
            )[groupby_fn_to_pandas_filter(df, filter_fn_node.from_op.inputs["key"])]
        elif op_name == "dict":
            # Return as list... though we'll need to keep track of that
            # we did this and remap the result keys.
            # TODO
            return list(
                groupby_fn_to_pandas_filter(df, n)
                for n in filter_fn_node.from_op.inputs.values()
            )
        raise Exception("unhandled op name", op_name)
    elif isinstance(filter_fn_node, graph.VarNode):
        if filter_fn_node.name == "row":
            return df
        raise Exception("unhandled var name")


def numpy_val_to_python(val):
    if isinstance(val, numpy.integer):
        return int(val)
    elif isinstance(val, float):
        if math.isnan(val):
            return "nan"
        return float(val)
    return val


@dataclasses.dataclass(frozen=True)
class DataFrameType(types.Type):
    instance_classes = pandas.DataFrame
    name = "dataframe"

    object_type: types.Type = types.Any()

    def __str__(self):
        return "<DataFrameType %s>" % self.object_type

    def _to_dict(self):
        return {"objectType": self.object_type.to_dict()}

    @classmethod
    def from_dict(cls, d):
        return cls(types.TypeRegistry.type_from_dict(d["objectType"]))

    @classmethod
    def type_of_instance(cls, obj):
        obj_prop_types = {}
        for col_name, dtype in obj.dtypes.items():
            if dtype == np.dtype("object"):
                weave_type = types.String()
            elif dtype == np.dtype("int64"):
                weave_type = types.Int()
            elif dtype == np.dtype("float64"):
                weave_type = types.Float()
            elif dtype == np.dtype("bool"):
                weave_type = types.Boolean()

            else:
                raise errors.WeaveTypeError(
                    "Type conversion not implemented for dtype: %s" % dtype
                )
            obj_prop_types[col_name] = weave_type
        return cls(types.TypedDict(obj_prop_types))

    def save_instance(self, obj, artifact, name):
        table = pa.Table.from_pandas(obj)
        with artifact.new_file(f"{name}.parquet", binary=True) as f:
            pq.write_table(table, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.parquet", binary=True) as f:
            table = pq.read_table(f)
        return table.to_pandas()


@dataclasses.dataclass(frozen=True)
class DataFrameTableType(types.Type):
    _base_type = types.List
    name = "dataframeTable"

    object_type: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj):
        return cls(obj.object_type)

    def save_instance(self, obj, artifact, name):
        d = {"_df": obj._df, "object_type": obj.object_type}
        type_of_d = types.TypedDict(
            {
                "_df": DataFrameType(),
                "object_type": types.TypeType(),
            }
        )

        serializer = mappers_python.map_to_python(type_of_d, artifact)
        result_d = serializer.apply(d)

        with artifact.new_file(f"{name}.DataFrameTable.json") as f:
            json.dump(result_d, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.DataFrameTable.json") as f:
            result = json.load(f)
        type_of_d = types.TypedDict(
            {
                "_df": DataFrameType(),
                "object_type": types.TypeType(),
            }
        )
        mapper = mappers_python.map_from_python(type_of_d, artifact)
        res = mapper.apply(result)
        return self.instance_class(artifact=artifact, **res)


def index_output_type(input_types):
    # THIS IS NO GOOD
    # TODO: need to fix Const type so we don't need this.
    self_type = input_types["self"]
    if isinstance(self_type, types.Const):
        return self_type.val_type.object_type
    else:
        return self_type.object_type


@weave_class(weave_type=DataFrameTableType)
class DataFrameTable:
    _df: pandas.DataFrame
    object_type: types.Type

    def __init__(self, _df, object_type=None, artifact=None):
        self._df = _df
        self.object_type = object_type
        if self.object_type is None:
            self.object_type = types.TypeRegistry.type_of(self._df).object_type

    def __iter__(self):
        # Iter is implemented to make list_indexCheckpoint work, which
        # currently expects to be able to iterate. Iteration gets stuck in an
        # infinite loop without implementing __iter__ here, since Python will
        # default to using __getitem__, which is a lazy Weave op.
        # TODO: When we fix arrow/vectorization and tagging, we'll have to
        # fix this.
        for row in self._df.itertuples():
            yield row._asdict()

    def _count(self):
        return len(self._df)

    def __len__(self):
        return self._count()

    @op()
    def count(self) -> int:
        return self._count()

    def _index(self, index):
        try:
            # TODO
            # to_json and back to dict then back to json :(
            return json.loads(self._df.iloc[index].to_json())
        except IndexError:
            return None

    @op(output_type=index_output_type)
    def __getitem__(self, index: int):
        return self._index(index)

    # @op(output_type=mapped_pick_output_type)
    # def pick(self, key: str):
    #     return self._df[key]

    @op(
        input_type={
            "filterFn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: input_types["self"],
    )
    def filter(self, filterFn):
        return DataFrameTable(self._df[filter_fn_to_pandas_filter(self._df, filterFn)])

    @op(
        input_type={
            "map_fn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(input_types["self"].object_type),
    )
    def map(self, map_fn):
        self_list = []
        for i in range(self._count()):
            self_list.append(self._index(i))
        return list_.List.map.resolve_fn(self_list, map_fn)

    @op(
        input_type={
            "group_by_fn": lambda input_types: types.Function(
                {"row": input_types["self"].object_type}, types.Any()
            ),
        },
        output_type=lambda input_types: types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict(
                    {
                        "groupKey": input_types["group_by_fn"].output_type,
                    }
                ),
                types.List(input_types["self"].object_type),
            )
        ),
    )
    def groupby(self, group_by_fn):
        group_keys = None
        if group_by_fn.from_op.name == "dict":
            group_keys = list(group_by_fn.from_op.inputs.keys())
        pandas_gb = groupby_fn_to_pandas_filter(self._df, group_by_fn)
        grouped = self._df.groupby(pandas_gb)
        df = grouped.agg(list)
        result = []
        for i in range(len(df)):
            row = df.iloc[i]
            row_result = []
            for row_val in zip(*row.values):
                row_result.append(dict(zip(df.keys(), row_val)))
            group_key_vals = row.name
            if isinstance(group_key_vals, tuple):
                group_key_vals = [numpy_val_to_python(v) for v in group_key_vals]
            else:
                group_key_vals = numpy_val_to_python(group_key_vals)
            if group_keys is None:
                group_key = group_key_vals
            else:
                if not isinstance(group_key_vals, tuple):
                    group_key_vals = (group_key_vals,)
                group_key = dict(zip(group_keys, group_key_vals))
            item = box.box(row_result)
            tag_store.add_tags(item, {"groupKey": group_key})
            result.append(item)
        return result


DataFrameTableType.instance_classes = DataFrameTable
DataFrameTableType.instance_class = DataFrameTable


def _pd_dtype_to_weave(dtype):
    if dtype.name == "object" or dtype.name == "category":
        return types.String()
    elif dtype.name == "int64":
        return types.Int()
    elif dtype.name == "float64":
        return types.Float()
    elif dtype.name == "bool":
        return types.Boolean()
    elif dtype.name == "datetime64" or dtype.name == "timedelta[ns]":
        return types.Timestamp()
    else:
        raise NotImplementedError("Unsupported dtype: {}".format(dtype))


@op(
    name="file-refine_pandasreadcsv",
    hidden=True,
    input_type={"file": file_base.FileBaseType()},
    output_type=types.TypeType(),
)
def refine_pandasreadcsv(file):
    res = pandasreadcsv.raw_resolve_fn(file)
    columns = res._df.dtypes.index.values.tolist()
    dtypes = res._df.dtypes.values.tolist()
    prop_types = {col: _pd_dtype_to_weave(dtype) for col, dtype in zip(columns, dtypes)}
    return DataFrameTableType(types.TypedDict(prop_types))


@op(
    name="file-pandasreadcsv",
    input_type={"file": file_base.FileBaseType()},
    output_type=DataFrameTableType(),
    refine_output_type=refine_pandasreadcsv,
)
def pandasreadcsv(file):
    local_path = file.get_local_path()
    # Warning, terrible hack to make demo work
    try:
        return DataFrameTable(pandas.read_csv(local_path, low_memory=False))
    except:
        return DataFrameTable(
            pandas.read_csv(local_path, delimiter=";", low_memory=False)
        )



import os

import pytest

import weave


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_mistral_quickstart(client: weave.trace.weave_client.WeaveClient) -> None:
    from mistralai.client import MistralClient

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = MistralClient(api_key=api_key)

    chat_response = mistral_client.chat(
        model=model,
        messages=[dict(role="user", content="What is the best French cheese?")],
    )

    all_content = chat_response.choices[0].message.content
    exp = """The "best" French cheese can vary greatly depending on personal preferences, as there are hundreds of different types of French cheeses, each with its unique flavor, texture, and aroma. However, some French cheeses are particularly popular and renowned:

1. Brie de Meaux: Often simply called Brie, this is a soft cheese with a white rind and a creamy, rich interior. It's one of the most well-known French cheeses internationally.

2. Camembert: Similar to Brie, Camembert is a soft, surface-ripened cheese. It has a stronger flavor and aroma compared to Brie.

3. Roquefort: This is a blue cheese made from sheep's milk. It's tangy, sharp, and slightly salty.

4. Comté: A hard cheese made from unpasteurized cow's milk, Comté has a nutty, slightly sweet flavor.

5. Reblochon: This is a soft, rind-washed cheese with a nutty and fruity taste. It's often used in tartiflette, a classic French dish from the Savoie region.

6. Époisses: Known for its pungent smell, Époisses is a soft, washed-rind cheese with a rich and creamy flavor."""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == chat_response.id
    assert output.model == chat_response.model
    assert output.object == chat_response.object
    assert output.created == chat_response.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"] == 299
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"] == 10
    assert output.usage.total_tokens == model_usage["total_tokens"] == 309


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_mistral_quickstart_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai.async_client import MistralAsyncClient

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = MistralAsyncClient(api_key=api_key)

    chat_response = await mistral_client.chat(
        model=model,
        messages=[dict(role="user", content="What is the best French cheese?")],
    )

    all_content = chat_response.choices[0].message.content
    exp = """There are many excellent French cheeses, and the "best" one often depends on personal preference. However, some of the most popular and highly regarded French cheeses include:

* Comté: A hard cheese made from unpasteurized cow's milk in the Franche-Comté region of eastern France. It has a rich, nutty flavor and a firm, slightly granular texture.
* Camembert: A soft, surface-ripened cheese made from cow's milk in Normandy. It has a bloomy white rind and a creamy, earthy flavor.
* Roquefort: A blue cheese made from sheep's milk in the south of France. It has a tangy, slightly salty flavor and a crumbly texture.
* Brie: A soft cheese made from cow's milk in the Île-de-France region around Paris. It has a white, edible rind and a creamy, buttery flavor.
* Reblochon: A soft, washed-rind cheese made from cow's milk in the Savoie region of the French Alps. It has a nutty, fruity flavor and a soft, supple texture.

Ultimately, the best French cheese is a matter of personal taste. I would recommend trying a variety of cheeses and seeing which ones you enjoy the most."""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == chat_response.id
    assert output.model == chat_response.model
    assert output.object == chat_response.object
    assert output.created == chat_response.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"] == 297
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"] == 10
    assert output.usage.total_tokens == model_usage["total_tokens"] == 307


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_mistral_quickstart_with_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai.client import MistralClient

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = MistralClient(api_key=api_key)

    chat_response = mistral_client.chat_stream(
        model=model,
        messages=[dict(role="user", content="What is the best French cheese?")],
    )

    all_content = ""
    for chunk in chat_response:
        all_content += chunk.choices[0].delta.content

    exp = """France is known for its diverse and high-quality cheeses, so the "best" French cheese can depend on personal preference. However, some of the most renowned French cheeses include:

1. Comté: A hard cheese made from unpasteurized cow's milk in the Franche-Comté region. It has a nutty, slightly sweet flavor.

2. Brie de Meaux: Often simply called Brie, this is a soft cheese with a white rind. It's known for its creamy texture and mild, slightly tangy flavor.

3. Roquefort: This is a blue cheese made from sheep's milk. It has a strong, tangy flavor and a crumbly texture.

4. Camembert: Similar to Brie, Camembert is a soft cheese with a white rind. However, it has a stronger, more earthy flavor.

5. Reblochon: A soft cheese from the Savoie region, it's known for its fruity and nutty taste with a slight bitterness.

6. Époisses: This is a pungent soft cheese with a distinctive orange rind. It's known for its strong flavor and creamy texture."""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == chunk.id
    assert output.model == chunk.model
    assert output.object == chunk.object
    assert output.created == chunk.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"] == 274
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"] == 10
    assert output.usage.total_tokens == model_usage["total_tokens"] == 284


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_mistral_quickstart_with_stream_async(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from mistralai.async_client import MistralAsyncClient

    # This is taken directly from https://docs.mistral.ai/getting-started/quickstart/
    api_key = os.environ.get("MISTRAL_API_KEY", "DUMMY_API_KEY")
    model = "mistral-large-latest"

    mistral_client = MistralAsyncClient(api_key=api_key)

    chat_response = mistral_client.chat_stream(
        model=model,
        messages=[dict(role="user", content="What is the best French cheese?")],
    )

    all_content = ""
    async for chunk in chat_response:
        all_content += chunk.choices[0].delta.content

    exp = """The "best" French cheese can depend on personal preferences, but here are a few popular ones:

1. Brie: Often referred to as "The Queen of Cheeses," Brie is a soft cheese named after the French region Brie. It has a mild, slightly sweet flavor with a creamy texture.

2. Camembert: This is another soft cheese from Normandy, France. It has a stronger flavor than Brie, with a hint of mushroom taste.

3. Roquefort: This is a blue cheese made from sheep's milk. It's aged in the natural Combalou caves of Roquefort-sur-Soulzon, which gives it a unique, tangy flavor.

4. Comté: This is a hard cheese made from unpasteurized cow's milk in the Franche-Comté region of eastern France. It has a complex, nutty flavor.

5. Chèvre: This is a general term for goat cheese in French. It can come in many forms and flavors, from fresh and mild to aged and tangy."""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.choices[0].message.content == exp
    assert output.choices[0].finish_reason == "stop"
    assert output.id == chunk.id
    assert output.model == chunk.model
    assert output.object == chunk.object
    assert output.created == chunk.created
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.completion_tokens == model_usage["completion_tokens"] == 242
    assert output.usage.prompt_tokens == model_usage["prompt_tokens"] == 10
    assert output.usage.total_tokens == model_usage["total_tokens"] == 252



import typing

import pyarrow as pa
import pyarrow.compute as pc

from weave.legacy.weave import api as api
from weave.legacy.weave import weave_internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import errors, arrow_util, artifact_base, artifact_mem, box, mappers_arrow
from weave.legacy.weave.arrow.arrow import (
    ArrowWeaveListType,
)
from weave.legacy.weave.arrow.list_ import (
    ArrowWeaveList,
    PathType,
    unsafe_awl_construction,
)
from weave.legacy.weave.language_features.tagging import tag_store, tagged_value_type

# Hmm... this doesn't work on ObjectType, which contains a Union of Struct...
# We need that because our ImageFileArtifactRefType has a union of structs
# in its property types.

# Could test a fix by hard-coding that to be the merged type, just to show it
# works...


def recursively_merge_union_types_if_they_are_unions_of_structs(
    type_: types.Type,
) -> types.Type:
    """Input preprocessor for to_arrow()."""
    if isinstance(type_, types.TypedDict):
        return types.TypedDict(
            {
                k: recursively_merge_union_types_if_they_are_unions_of_structs(v)
                for k, v in type_.property_types.items()
            }
        )
    elif isinstance(type_, types.UnionType):
        if type_.is_simple_nullable() or len(type_.members) < 2:
            return type_

        new_type = type_.members[0]
        for member in type_.members[1:]:
            new_type = types.merge_types(new_type, member)

        if isinstance(new_type, types.UnionType):
            # cant go down any further
            return new_type

        return recursively_merge_union_types_if_they_are_unions_of_structs(new_type)

    elif isinstance(type_, types.List):
        return types.List(
            recursively_merge_union_types_if_they_are_unions_of_structs(
                type_.object_type
            )
        )
    elif isinstance(type_, ArrowWeaveListType):
        return ArrowWeaveListType(
            recursively_merge_union_types_if_they_are_unions_of_structs(
                type_.object_type
            )
        )
    elif isinstance(type_, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueType(
            typing.cast(
                types.TypedDict,
                recursively_merge_union_types_if_they_are_unions_of_structs(type_.tag),
            ),
            recursively_merge_union_types_if_they_are_unions_of_structs(type_.value),
        )

    return type_


def recursively_build_pyarrow_array(
    py_objs: list[typing.Any],
    pyarrow_type: typing.Union[pa.DataType, arrow_util.ArrowTypeWithFieldInfo],
    mapper,
    py_objs_already_mapped: bool = False,
) -> pa.Array:
    arrays: list[pa.Array] = []

    if isinstance(pyarrow_type, arrow_util.ArrowTypeWithFieldInfo):
        pyarrow_type = arrow_util.arrow_type(pyarrow_type)
    pyarrow_type = typing.cast(pa.DataType, pyarrow_type)

    def none_unboxer(iterator: typing.Iterable):
        for obj in iterator:
            if isinstance(obj, box.BoxedNone):
                # get rid of box
                yield None
            else:
                yield obj

    if isinstance(mapper.type, types.UnionType) and mapper.type.is_simple_nullable():
        nonnull_mapper = [
            m for m in mapper._member_mappers if m.type != types.NoneType()
        ][0]

        return recursively_build_pyarrow_array(
            list(none_unboxer(py_objs)),
            pyarrow_type,
            nonnull_mapper,
            py_objs_already_mapped,
        )
    elif pa.types.is_null(pyarrow_type):
        return pa.array(
            none_unboxer(py_objs),
            type=pyarrow_type,
        )

    elif pa.types.is_struct(pyarrow_type):
        keys: list[str] = []
        # keeps track of null values so that we can null entries at the struct level
        mask: list[bool] = []

        assert isinstance(
            mapper,
            (
                mappers_arrow.TypedDictToArrowStruct,
                mappers_arrow.TaggedValueToArrowStruct,
                mappers_arrow.ObjectToArrowStruct,
                mappers_arrow.GQLHasKeysToArrowStruct,
            ),
        )

        if isinstance(mapper, mappers_arrow.GQLHasKeysToArrowStruct):
            _dictionary: dict[typing.Optional[str], typing.Tuple[int, typing.Any]] = {}
            indices: list[typing.Optional[int]] = []
            for py_obj in py_objs:
                id = py_obj["id"] if py_obj is not None else None
                if id in _dictionary:
                    indices.append(_dictionary[id][0])
                else:
                    new_index = len(_dictionary)
                    _dictionary[id] = (new_index, py_obj)
                    indices.append(new_index)

            dictionary = [
                mapper.apply(v[1]) if v[1] is not None else None
                for v in _dictionary.values()
            ]
            array = recursively_build_pyarrow_array(
                dictionary,
                pyarrow_type,
                mapper.as_typeddict_mapper(),
                True,
            )

            indices = pa.array(indices, type=pa.int32())
            return pa.DictionaryArray.from_arrays(indices, array)

        else:
            # handle empty struct case - the case where the struct has no fields
            if len(pyarrow_type) == 0:
                return pa.array(py_objs, type=pyarrow_type)

            for i, field in enumerate(pyarrow_type):
                data: list[typing.Any] = []
                if isinstance(
                    mapper,
                    mappers_arrow.TypedDictToArrowStruct,
                ):
                    for py_obj in py_objs:
                        if py_obj is None:
                            data.append(None)
                        else:
                            data.append(py_obj.get(field.name, None))
                        if i == 0:
                            mask.append(py_obj is None)

                    array = recursively_build_pyarrow_array(
                        data,
                        field.type,
                        mapper._property_serializers[field.name],
                        py_objs_already_mapped,
                    )
                elif isinstance(
                    mapper,
                    mappers_arrow.ObjectToArrowStruct,
                ):
                    for py_obj in py_objs:
                        if py_obj is None:
                            data.append(None)
                        elif py_objs_already_mapped:
                            data.append(py_obj.get(field.name, None))
                        else:
                            data.append(getattr(py_obj, field.name, None))
                        if i == 0:
                            mask.append(py_obj is None)

                    array = recursively_build_pyarrow_array(
                        data,
                        field.type,
                        mapper._property_serializers[field.name],
                        py_objs_already_mapped,
                    )

                elif isinstance(mapper, mappers_arrow.TaggedValueToArrowStruct):
                    if field.name == "_tag":
                        for py_obj in py_objs:
                            if py_obj is None:
                                data.append(None)
                            else:
                                data.append(tag_store.get_tags(py_obj))
                            if i == 0:
                                mask.append(py_obj is None)

                        array = recursively_build_pyarrow_array(
                            data,
                            field.type,
                            mapper._tag_serializer,
                            py_objs_already_mapped,
                        )
                    else:
                        for py_obj in py_objs:
                            if py_obj is None:
                                data.append(None)
                            else:
                                data.append(box.unbox(py_obj))
                            if i == 0:
                                mask.append(py_obj is None)

                        array = recursively_build_pyarrow_array(
                            data,
                            field.type,
                            mapper._value_serializer,
                            py_objs_already_mapped,
                        )
                else:
                    assert isinstance(
                        mapper,
                        mappers_arrow.GQLHasKeysToArrowStruct,
                    )
                    for py_obj in py_objs:
                        if py_obj is None:
                            data.append(None)
                        else:
                            data.append(mapper.apply(py_obj))
                        if i == 0:
                            mask.append(py_obj is None)

                    array = recursively_build_pyarrow_array(
                        data,
                        field.type,
                        mapper._property_serializers[field.name],
                        False,
                    )

                arrays.append(array)
                keys.append(field.name)
            return pa.StructArray.from_arrays(
                arrays, keys, mask=pa.array(mask, type=pa.bool_())
            )
    elif pa.types.is_union(pyarrow_type):
        assert isinstance(mapper, mappers_arrow.UnionToArrowUnion)
        type_codes: list[int] = [
            0 if o == None else mapper.type_code_of_obj(o) for o in py_objs
        ]
        offsets: list[int] = []
        py_data: list[list] = []
        for _ in range(len(pyarrow_type)):
            py_data.append([])

        for row_index, type_code in enumerate(type_codes):
            offsets.append(len(py_data[type_code]))
            py_data[type_code].append(py_objs[row_index])

        for i, raw_py_data in enumerate(py_data):
            array = recursively_build_pyarrow_array(
                raw_py_data,
                pyarrow_type.field(i).type,
                mapper.mapper_of_type_code(i),
                py_objs_already_mapped,
            )
            arrays.append(array)

        return pa.UnionArray.from_dense(
            pa.array(type_codes, type=pa.int8()),
            pa.array(offsets, type=pa.int32()),
            arrays,
        )
    elif pa.types.is_list(pyarrow_type):
        assert isinstance(mapper, mappers_arrow.ListToArrowArr)
        offsets = [0]
        flattened_objs = []
        mask = []
        for obj in py_objs:
            mask.append(obj == None)
            if obj == None:
                obj = []
            offsets.append(offsets[-1] + len(obj))
            flattened_objs += obj
        new_objs = recursively_build_pyarrow_array(
            flattened_objs,
            pyarrow_type.value_type,
            mapper._object_type,
            py_objs_already_mapped,
        )
        return pa.ListArray.from_arrays(
            offsets, new_objs, mask=pa.array(mask, type=pa.bool_())
        )
    elif pa.types.is_temporal(pyarrow_type):
        if py_objs_already_mapped:
            return pa.array(py_objs, type=pyarrow_type)

    values = [mapper.apply(o) if o is not None else None for o in py_objs]

    # These are plain values.

    if mapper.type == types.Number():
        # Let pyarrow infer this type.
        # This covers the case where a Weave0 table includes a Number column that
        # contains integers that are too large for float64. We map Number to float64,
        # but allow it to be int64 as well in our ArrowWeaveList.validate method.
        res = pa.array(values)
        if pa.types.is_null(res.type):
            res = res.cast(pa.int64())
    else:
        res = pa.array(values, type=pyarrow_type)
    return res


# This will be a faster version fo to_arrow (below). Its
# used in op file-table, to convert from a wandb Table to Weave
# (that code is very experimental and not totally working yet)
def to_arrow_from_list_and_artifact(
    obj: typing.Any,
    object_type: types.Type,
    artifact: artifact_base.Artifact,
    py_objs_already_mapped: bool = False,
) -> ArrowWeaveList:
    # Get what the parquet type will be.
    merged_object_type = recursively_merge_union_types_if_they_are_unions_of_structs(
        object_type
    )
    mapper = mappers_arrow.map_to_arrow(merged_object_type, artifact)
    pyarrow_type = mapper.result_type()

    arrow_obj = recursively_build_pyarrow_array(
        obj, pyarrow_type, mapper, py_objs_already_mapped=py_objs_already_mapped
    )
    return ArrowWeaveList(arrow_obj, merged_object_type, artifact)


def to_arrow(
    obj,
    wb_type=None,
    artifact: typing.Optional[artifact_base.Artifact] = None,
):
    if isinstance(obj, ArrowWeaveList):
        return obj
    if wb_type is None:
        wb_type = types.type_of_with_refs(obj)
    artifact = artifact or artifact_mem.MemArtifact()
    outer_tags: typing.Optional[dict[str, typing.Any]] = None
    if isinstance(wb_type, tagged_value_type.TaggedValueType):
        outer_tags = tag_store.get_tags(obj)
        wb_type = wb_type.value
    if isinstance(wb_type, types.List):
        merged_object_type = (
            recursively_merge_union_types_if_they_are_unions_of_structs(
                wb_type.object_type
            )
        )

        # Convert to arrow, serializing Custom objects to the artifact
        mapper = mappers_arrow.map_to_arrow(merged_object_type, artifact)
        pyarrow_type = arrow_util.arrow_type(mapper.result_type())

        arrow_obj = recursively_build_pyarrow_array(obj, pyarrow_type, mapper)
        weave_obj: ArrowWeaveList = ArrowWeaveList(
            arrow_obj, merged_object_type, artifact
        )

        # Save the weave object to the artifact
        # ref = storage.save(weave_obj, artifact=artifact)
        if outer_tags is not None:
            tag_store.add_tags(weave_obj, outer_tags)

        return weave_obj

    raise errors.WeaveInternalError("to_arrow not implemented for: %s" % wb_type)


def to_weave_arrow(v: typing.Any):
    awl = to_arrow(v)
    return weave_internal.const(awl)


def to_parquet_friendly(l: ArrowWeaveList) -> ArrowWeaveList:
    def _convert_col_to_parquet_friendly(
        col: ArrowWeaveList, path: PathType
    ) -> typing.Optional[ArrowWeaveList]:
        _, non_none_type = types.split_none(col.object_type)
        if isinstance(non_none_type, types.UnionType):
            full_length_fields = []
            for i in range(len(non_none_type.members)):
                member_field = col._arrow_data.field(i)
                padding = len(col) - len(member_field)
                if padding > 0:
                    member_field = pa.concat_arrays(
                        [member_field, pa.nulls(padding, type=member_field.type)]
                    )
                else:
                    # A case where this can happen:
                    # pa.nulls(1, pa.list_(pa.dense_union([pa.field('a', pa.int32()), pa.field('b', pa.float64())])))
                    # Then union in the list has type_codes and offsets of len zero, but a field
                    # of len one that has the null.
                    member_field = member_field.slice(0, len(col))
                full_length_fields.append(member_field)
            struct = pa.StructArray.from_arrays(
                [col._arrow_data.type_codes, col._arrow_data.offsets]
                + full_length_fields,
                ["type_codes", "offsets"]
                + [str(i) for i in list(range(len(full_length_fields)))],
            )
            return ArrowWeaveList(struct, col.object_type, col._artifact)
        elif isinstance(col.object_type, types.TypedDict):
            if not col.object_type.property_types:
                return ArrowWeaveList(
                    col._arrow_data.is_valid(),
                    col.object_type,
                    col._artifact,
                )
        return None

    with unsafe_awl_construction("to_parquet_friendly"):
        return l.map_column(_convert_col_to_parquet_friendly)


def from_parquet_friendly(l: ArrowWeaveList) -> ArrowWeaveList:
    def _ident(col: ArrowWeaveList, path: PathType) -> typing.Optional[ArrowWeaveList]:
        return None

    def _convert_col_from_parquet_friendly(
        col: ArrowWeaveList, path: PathType
    ) -> typing.Optional[ArrowWeaveList]:
        _, non_none_type = types.split_none(col.object_type)
        if isinstance(non_none_type, types.UnionType):
            struct = col._arrow_data
            type_codes = struct.field("type_codes")
            offsets = struct.field("offsets")
            member_fields = []
            for i in range(len(non_none_type.members)):
                full_length_field = struct.field(str(i))
                mask = pa.compute.equal(type_codes, i).cast(pa.int32())
                masked_offsets = pa.compute.multiply(offsets, mask)
                max_offset = pa.compute.max(masked_offsets).as_py()
                if max_offset is None:
                    max_offset = 0
                else:
                    max_offset += 1

                # TODO: This will hold all the zero memory!
                member_field = full_length_field.slice(0, max_offset)
                member_fields.append(member_field)

            if type_codes.null_count > 0 or offsets.null_count > 0:
                if type_codes.null_count != len(type_codes):
                    # I think this only happens when we have empty objects somewhere
                    # in the tree?
                    raise errors.WeaveInternalError(
                        "Unexpected nulls in type_codes for union"
                    )
                type_codes = type_codes.fill_null(0)
                offsets = offsets.fill_null(0)
                if len(member_fields[0]) == 0:
                    member_fields[0] = pa.nulls(1, member_fields[0].type)
            union = pa.UnionArray.from_dense(
                type_codes.fill_null(0), offsets.fill_null(0), member_fields
            )
            return ArrowWeaveList(union, col.object_type, col._artifact)
        elif isinstance(col.object_type, types.TypedDict):
            if not col.object_type.property_types:
                # Can have nulls because map_columns can put masks on parents in the
                # to_parquet pass.
                valid = col._arrow_data.fill_null(False)
                return ArrowWeaveList(
                    pa.array(
                        [{}] * len(col._arrow_data),
                        mask=pa.compute.invert(valid),
                        type=pa.struct([]),
                    ),
                    col.object_type,
                    col._artifact,
                )
        return None

    return l.map_column(_ident, _convert_col_from_parquet_friendly)


def simple_to_string(arr: pa.Array):
    return pa.compute.binary_join_element_wise(
        "__t_%s" % arr.type.id,
        arr.cast(pa.string()),
        "-",
    )


def to_compare_safe(awl: ArrowWeaveList) -> ArrowWeaveList:
    """Converts any ArrowWeaveList to simple type that pa.compute.equal can compare."""
    from weave.legacy.weave.ops_domain.wbmedia import ArtifactAssetType

    # Returns a number of string arrow weave list, possibly with Nones
    def _to_compare_safe(
        col: ArrowWeaveList, path: PathType
    ) -> typing.Optional[ArrowWeaveList]:
        if pa.types.is_null(col._arrow_data.type):
            return ArrowWeaveList(
                pa.nulls(len(col), type=pa.string()),
                types.String(),
                None,
            )
        elif pa.types.is_string(col._arrow_data.type):
            return ArrowWeaveList(col._arrow_data, types.String(), None)
        elif pa.types.is_dictionary(col._arrow_data.type):
            if not isinstance(col.object_type, types.String):
                raise errors.WeaveInternalError(
                    "Unexpected dictionary type for non-string type"
                )
            return ArrowWeaveList(col._arrow_data, types.String(), None)
        elif pa.types.is_floating(col._arrow_data.type):
            # Ensure that -0.0 is 0. If we end up converting to a string
            # later (which happens if we have non-numeric types within a union)
            # then -0.0 will be converted to "-0.0" which is not equal to "0.0"
            return ArrowWeaveList(
                pc.choose(pc.equal(col._arrow_data, 0), col._arrow_data, 0),
                types.Number(),
                None,
            )
        elif pa.types.is_integer(col._arrow_data.type):
            return ArrowWeaveList(col._arrow_data, types.Number(), None)
        elif pa.types.is_timestamp(col._arrow_data.type):
            # Cast to int64 and then string. Leaving this as timestamp
            # means it will later be cast directly to string, which is very expensive in
            # pyarrow (1.6s for 500k records v. 0.01s for int64, a 100x improvement)
            return ArrowWeaveList(
                col._arrow_data.cast(pa.int64()).cast(pa.string()), types.String(), None
            )
        elif pa.types.is_boolean(col._arrow_data.type):
            return ArrowWeaveList(col._arrow_data, types.Boolean(), None)
        elif ArtifactAssetType.assign_type(col.object_type):
            # Special logic as implemented in Weave0 for media types that contain sha256
            # checksums.
            # This idea is generalized already in Weave1. Any persistent ref to an artifact
            # is a uniquely comparable string for that object.
            return ArrowWeaveList(col._arrow_data.field("sha256"), types.String(), None)
        elif pa.types.is_struct(col._arrow_data.type):
            value_string_arrs = []
            field_names = sorted(field.name for field in col._arrow_data.type)
            for field_name in field_names:
                value_string_arrs.append(
                    pa.compute.binary_join_element_wise(
                        "__sk_%s" % (field_name),
                        simple_to_string(col._arrow_data.field(field_name)).fill_null(
                            "__none_"
                        ),
                        "_",
                    )
                )
            if not value_string_arrs:
                return ArrowWeaveList(
                    pa.nulls(len(col), type=pa.string()),
                    types.String(),
                    None,
                )
            struct_strings = pa.compute.binary_join_element_wise(
                *value_string_arrs, "-"
            )
            return ArrowWeaveList(
                pa.compute.replace_with_mask(
                    struct_strings,
                    pa.compute.invert(col._arrow_data.is_valid()),
                    pa.nulls(len(col._arrow_data), pa.string()),
                ),
                types.String(),
                None,
            )
        elif pa.types.is_list(col._arrow_data.type):
            stringed_list = pa.ListArray.from_arrays(
                col._arrow_data.offsets,
                simple_to_string(col._arrow_data.values).fill_null("__none_"),
            )
            list_strings = pa.compute.binary_join_element_wise(
                "__list_",
                pa.compute.binary_join(stringed_list, "-"),
                "-",
            )
            res: ArrowWeaveList = ArrowWeaveList(
                pa.compute.replace_with_mask(
                    list_strings,
                    pa.compute.invert(col._arrow_data.is_valid()),
                    pa.nulls(len(col._arrow_data), pa.string()),
                ),
                types.String(),
                None,
            )
            return res
        elif pa.types.is_union(col._arrow_data.type):
            merged = pa.nulls(len(col), pa.string())
            for type_code in range(len(col._arrow_data.type)):
                field = col._arrow_data.field(type_code)
                if len(field) == 0:
                    continue
                string_field = simple_to_string(field)
                mask = pa.compute.equal(col._arrow_data.type_codes, type_code)
                indexes = pa.compute.multiply(
                    mask.cast(pa.int8()), col._arrow_data.offsets
                )
                values = string_field.take(indexes)
                merged = pa.compute.if_else(mask, values, merged)

            return ArrowWeaveList(
                merged,
                types.String(),
                None,
            )
        else:
            raise errors.WeaveInternalError(
                'Unhandled type in "to_compare_safe" %s' % col._arrow_data.type
            )

    return awl.map_column(_to_compare_safe)


def unify_types(*arrs: ArrowWeaveList):
    """Ensures each arr has the same type using merge_types/concat."""
    # We make use of concat, which converts its inputs to the same type.
    if not arrs:
        return ()
    concatted = arrs[0]
    for a in arrs[1:]:
        concatted = concatted.concat(a)
    result = []
    arr_start_index = 0
    for a in arrs:
        result.append(concatted._slice(arr_start_index, arr_start_index + len(a)))
        arr_start_index += len(a)
    return result



import dataclasses

from weave.legacy.weave import api as weave
from weave.legacy.weave import weave_types as types


class MarkdownType(types.Type):
    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.md") as f:
            f.write(obj.md)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.md") as f:
            return Markdown(f.read())


@weave.weave_class(weave_type=MarkdownType)
@dataclasses.dataclass
class Markdown:
    md: str


MarkdownType.instance_classes = Markdown



import weave
from weave.legacy.weave import storage
from weave.legacy.weave import weave_types as types


def test_reffed_type(ref_tracking):
    obj_ref = storage.save([1, 2, 3])
    obj = obj_ref.get()
    obj_type = types.type_of_with_refs(obj)
    assert obj_type == types.LocalArtifactRefType(types.List(types.Int()))


def test_save_reffed_obj(ref_tracking):
    obj_ref = storage.save([1, 2, 3])
    obj = storage.get(str(obj_ref))
    obj_ref2 = storage.save(obj, "again")
    # Is this what we want or should it save?
    assert str(obj_ref) == str(obj_ref2)


def test_save_nested_reffed_obj(ref_tracking):
    obj_ref = storage.save([1, 2, 3])
    obj = obj_ref.get()
    outer_obj = {"a": obj}
    outer_ref = storage.save(outer_obj)
    outer_obj2 = storage.get(str(outer_ref))
    obj2 = outer_obj2["a"]
    obj2_ref = storage.get_ref(obj2)
    assert str(obj_ref) == str(obj2_ref)


def test_nested_object_deref(ref_tracking):
    @weave.type()
    class TestTypeA:
        val: int

    @weave.type()
    class TestTypeB:
        val: int
        a: TestTypeA

    a_ref = storage.save(TestTypeA(5))
    b_ref = storage.save(TestTypeB(6, a_ref))

    b = storage.get(str(b_ref))
    assert b.val == 6
    assert b.a.val == 5



from weave.legacy.weave import api as weave

from ...legacy.weave import registry_mem


def test_mapped_add():
    mapped_add = registry_mem.memory_registry.get_op("mapped_number-add")
    assert weave.use(mapped_add([3, 4, 5], 9)) == [12, 13, 14]



import importlib

from weave.trace.patcher import SymbolPatcher


def test_symbol_patcher():
    from .test_patcher_module.example_class import ExampleClass

    patcher = SymbolPatcher(
        lambda: importlib.import_module(
            "weave.trace.tests.test_patcher_module.example_class"
        ),
        "ExampleClass.example_fn",
        lambda original_fn: lambda self: 43,
    )

    obj = ExampleClass()
    assert obj.example_fn() == 42
    patcher.attempt_patch()
    obj = ExampleClass()
    assert obj.example_fn() == 43
    patcher.undo_patch()
    obj = ExampleClass()
    assert obj.example_fn() == 42


def test_symbol_patcher_invalid_module():
    patcher = SymbolPatcher(
        lambda: importlib.import_module("INVALID_MODULE"),
        "ExampleClass.example_fn",
        lambda original_fn: lambda self: 43,
    )
    # Should not raise, but should fail to patch
    assert not patcher.attempt_patch()
    assert not patcher.undo_patch()


def test_symbol_patcher_invalid_attr():
    patcher = SymbolPatcher(
        lambda: importlib.import_module(
            "weave.trace.tests.test_patcher_module.example_class"
        ),
        "NotARealExampleClass.example_fn",
        lambda original_fn: lambda self: 43,
    )
    # Should not raise, but should fail to patch
    assert not patcher.attempt_patch()
    assert not patcher.undo_patch()


def test_symbol_patcher_invalid_patching():
    from .test_patcher_module.example_class import ExampleClass

    patcher = SymbolPatcher(
        lambda: importlib.import_module(
            "weave.trace.tests.test_patcher_module.example_class"
        ),
        "ExampleClass.example_fn",
        lambda original_fn: [] + 42,
    )
    # Should not raise, but should fail to patch
    assert not patcher.attempt_patch()
    assert not patcher.undo_patch()
    obj = ExampleClass()
    assert obj.example_fn() == 42



import importlib
import typing
from functools import wraps

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from cohere.types.non_streamed_chat_response import NonStreamedChatResponse
    from cohere.v2.types.non_streamed_chat_response2 import NonStreamedChatResponse2


def cohere_accumulator(
    acc: typing.Optional[dict],
    value: typing.Any,
) -> "NonStreamedChatResponse":
    # don't need to accumulate, is build-in by cohere!
    # https://docs.cohere.com/docs/streaming
    # A stream-end event is the final event of the stream, and is returned only when streaming is finished.
    # This event contains aggregated data from all the other events such as the complete text,
    # as well as a finish_reason for why the stream ended (i.e. because of it was finished or there was an error).
    if acc is None:
        acc = {}

    # we wait for the last event
    if hasattr(value, "event_type"):
        if value.event_type == "stream-end" and value.is_finished:
            if value.response:
                acc = value.response
    return acc


def cohere_accumulator_v2(
    acc: typing.Optional[dict],
    value: typing.Any,
) -> "NonStreamedChatResponse2":
    from cohere.v2.types.assistant_message_response import AssistantMessageResponse
    from cohere.v2.types.non_streamed_chat_response2 import NonStreamedChatResponse2

    def _accumulate_content(
        prev: str,
        content: str,
    ) -> str:
        if isinstance(prev, str) and isinstance(content, str):
            prev += content
        return prev

    if acc is None:
        acc = NonStreamedChatResponse2(
            id=value.id,
            finish_reason=None,
            prompt=None,
            message=AssistantMessageResponse(
                role=value.delta.message.role,
                tool_calls=value.delta.message.tool_calls,
                tool_plan=value.delta.message.tool_plan,
                content=value.delta.message.content,
                citations=value.delta.message.citations,
            ),
            usage=None,
        )

    if value is None:
        return acc

    if value.type == "content-start" and value.delta.message.content.type == "text":
        if len(acc.message.content) == value.index:  # type: ignore
            acc.message.content.append(value.delta.message.content.text)  # type: ignore

    if value.type == "content-delta":
        _content = _accumulate_content(
            acc.message.content[value.index],  # type: ignore
            value.delta.message.content.text,  # type: ignore
        )
        acc.message.content[value.index] = _content  # type: ignore

    if value.type == "message-end":
        acc = acc.copy(  # type: ignore
            update={
                "finish_reason": value.delta.finish_reason,
                "usage": value.delta.usage,
            }
        )

    return acc


def cohere_wrapper(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return op

    return wrapper


def cohere_wrapper_v2(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        def _post_process_response(fn: typing.Callable) -> typing.Any:
            @wraps(fn)
            def _wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                response = fn(*args, **kwargs)

                try:
                    from cohere.v2.types.non_streamed_chat_response2 import (
                        NonStreamedChatResponse2,
                    )
                    from cohere.v2.types.usage import Usage

                    # Create a new instance with modified `usage`
                    response_dict = response.dict()
                    response_dict["usage"] = Usage(
                        billed_units=response.model_extra["meta"]["billed_units"],
                        tokens=response.model_extra["meta"]["tokens"],
                    )
                    response = NonStreamedChatResponse2(**response_dict)
                except:
                    pass  # prompt to upgrade cohere sdk

                return response

            return _wrapper

        op = weave.op()(_post_process_response(fn))
        op.name = name  # type: ignore
        return op

    return wrapper


def cohere_wrapper_async_v2(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        def _post_process_response(fn: typing.Callable) -> typing.Any:
            @wraps(fn)
            async def _wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                response = await fn(*args, **kwargs)

                try:
                    from cohere.v2.types.non_streamed_chat_response2 import (
                        NonStreamedChatResponse2,
                    )
                    from cohere.v2.types.usage import Usage

                    # Create a new instance with modified `usage`
                    response_dict = response.dict()
                    response_dict["usage"] = Usage(
                        billed_units=response.model_extra["meta"]["billed_units"],
                        tokens=response.model_extra["meta"]["tokens"],
                    )
                    response = NonStreamedChatResponse2(**response_dict)
                except:
                    pass  # prompt to upgrade cohere sdk

                return response

            return _wrapper

        op = weave.op()(_post_process_response(fn))
        op.name = name  # type: ignore
        return op

    return wrapper


def cohere_stream_wrapper(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return add_accumulator(op, lambda inputs: cohere_accumulator)  # type: ignore

    return wrapper


def cohere_stream_wrapper_v2(name: str) -> typing.Callable:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        op = weave.op()(fn)
        op.name = name  # type: ignore
        return add_accumulator(
            op, make_accumulator=lambda inputs: cohere_accumulator_v2
        )

    return wrapper


cohere_patcher = MultiPatcher(
    [
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "Client.chat",
            cohere_wrapper("cohere.Client.chat"),
        ),
        # Patch the async chat method
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "AsyncClient.chat",
            cohere_wrapper("cohere.AsyncClient.chat"),
        ),
        # Add patch for chat_stream method
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "Client.chat_stream",
            cohere_stream_wrapper("cohere.Client.chat_stream"),
        ),
        # Add patch for async chat_stream method
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "AsyncClient.chat_stream",
            cohere_stream_wrapper("cohere.AsyncClient.chat_stream"),
        ),
        # Add patch for cohere v2
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "ClientV2.chat",
            cohere_wrapper_v2("cohere.ClientV2.chat"),
        ),
        # Add patch for cohre v2 async chat method
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "AsyncClientV2.chat",
            cohere_wrapper_async_v2("cohere.AsyncClientV2.chat"),
        ),
        # Add patch for chat_stream method v2
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "ClientV2.chat_stream",
            cohere_stream_wrapper_v2("cohere.ClientV2.chat_stream"),
        ),
        # Add patch for async chat_stream method v2
        SymbolPatcher(
            lambda: importlib.import_module("cohere"),
            "AsyncClientV2.chat_stream",
            cohere_stream_wrapper_v2("cohere.AsyncClientV2.chat_stream"),
        ),
    ]
)



import binascii
import dataclasses
import io

import PIL
import PIL.Image

from weave.legacy.weave import api as weave
from weave.legacy.weave import weave_types as types


class ImageType(types.Type):
    name = "image"


@dataclasses.dataclass(frozen=True)
class PILImageType(types.Type):
    name = "pil_image"
    instance_classes = PIL.Image.Image

    width: types.Type = types.Int()
    height: types.Type = types.Int()
    mode: types.Type = types.String()

    @classmethod
    def type_of_instance(cls, obj: PIL.Image.Image):
        return cls(
            types.Const(types.Int(), obj.width),
            types.Const(types.Int(), obj.height),
            types.Const(types.String(), obj.mode),
        )

    @classmethod
    def save_instance(cls, obj: PIL.Image.Image, artifact, name):
        with artifact.new_file(f"{name}.png", binary=True) as f:
            obj.save(f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.png", binary=True) as f:
            im = PIL.Image.open(f)
            im.load()
            return im


@weave.weave_class(weave_type=PILImageType)
class PILImageOps:
    # TODO: should not need to hardcode type constants!
    @weave.op(input_type={"self": PILImageType()})
    def image_bytes(self) -> str:
        f = io.BytesIO()
        self.save(f, format="png", compress_level=3)  # type: ignore
        f.seek(0)
        res = binascii.hexlify(f.read()).decode("ISO-8859-1")
        return res

    @weave.op()
    def width_(self) -> int:
        return self.width  # type: ignore



import contextlib
import contextvars
import datetime
import functools
import os
import pathlib
import re
import typing

from weave.trace.client_context import weave_client as weave_client_context
from weave.legacy.weave import (
    artifact_base,
    artifact_fs,
    artifact_local,
    artifact_mem,
    artifact_wandb,
    box,
    errors,
    graph,
    mappers_python,
    ref_base,
    timestamp,
)

from weave.legacy.weave import weave_types as types

Ref = ref_base.Ref

if typing.TYPE_CHECKING:
    from weave.legacy.weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun


def split_path_dotfile(path, dotfile_name):  # type: ignore
    while path != "/":
        path, tail = os.path.split(path)
        if os.path.exists(os.path.join(path, dotfile_name)):
            return path, tail
    raise FileNotFoundError


def _get_name(wb_type: types.Type, obj: typing.Any) -> str:
    if getattr(obj, "name", None) != None:
        return obj.name
    return wb_type.name
    # This tries to figure out which variable references obj.
    # But it is slow when there are a lot of references. If we want to do
    # something like this, we'll need to do it somewhere closer to user
    # interaction.
    # obj_names = util.find_names(obj)
    # return f"{wb_type.name}-{obj_names[-1]}"


def _get_weave_type_with_refs(obj: typing.Any):  # type: ignore
    try:
        return types.type_of_with_refs(obj)
    except errors.WeaveTypeError as e:
        raise errors.WeaveSerializeError(
            "weave type error during serialization for object: %s. %s"
            % (obj, str(e.args))
        )


def _ensure_object_components_are_published(  # type: ignore
    obj: typing.Any, wb_type: types.Type, artifact: artifact_wandb.WandbArtifact
):
    from weave.legacy.weave.mappers_publisher import map_to_python_remote

    mapper = map_to_python_remote(wb_type, artifact)
    return mapper.apply(obj)


def _assert_valid_name_part(part: typing.Optional[str] = None):  # type: ignore
    if part is None:
        return
    if not re.match(r"^[a-zA-Z0-9_\-]+$", part):
        raise ValueError(
            "Invalid name part %s. Must be alphanumeric, dashes, or underscores." % part
        )


def _assert_valid_entity_name(part: typing.Optional[str] = None):  # type: ignore
    if part is None:
        return
    if not re.match(r"^[a-z0-9_\-]+$", part):
        raise ValueError(
            "Invalid entity name %s. Must be lowercase, digits, dashes, or underscores."
            % part
        )


def _assert_valid_project_name(part: typing.Optional[str] = None):  # type: ignore
    if part is None:
        return
    if len(part) > 128:
        raise ValueError("Invalid project name %s. Must be <= 128 characters." % part)
    if re.match(r"[\\#?%:]", part):
        raise ValueError(
            "Invalid project name %s. Can not contain \\, #, ?, %%, or :" % part
        )


def _assert_valid_artifact_name(part: typing.Optional[str] = None):  # type: ignore
    if part is None:
        return
    if len(part) > 128:
        raise ValueError("Invalid artifact name %s. Must be <= 128 characters." % part)
    if not re.match(r"^[a-zA-Z0-9_\-.]+$", part):  # from W&B Artifacts
        raise ValueError(
            "Invalid artifact name %s. Must be alphanumeric, periods, dashes, or underscores."
            % part
        )


class PublishTargetProject(typing.TypedDict):
    # TODO: should include entity_name as well, but we defalt that right now...
    # entity_name: str
    project_name: str


_pubish_target_project: contextvars.ContextVar[
    typing.Optional[typing.Optional[PublishTargetProject]]
] = contextvars.ContextVar("publish_target_project", default=None)


@contextlib.contextmanager
def publish_target_project(project: PublishTargetProject):  # type: ignore
    token = _pubish_target_project.set(project)
    yield
    _pubish_target_project.reset(token)


def get_publish_target_project() -> typing.Optional[PublishTargetProject]:
    return _pubish_target_project.get()


# Keep a cache of published local ref -> wandb ref. This is used to ensure
# we publish any needed op defs at the beginning of graph execution one time,
# to avoid parallel publishing them many times later in mapped operations.
PUBLISH_CACHE_BY_LOCAL_ART: dict[
    artifact_local.LocalArtifactRef, artifact_wandb.WandbArtifactRef
] = {}


def _direct_publish(
    obj: typing.Any,
    name: typing.Optional[str] = None,
    wb_project_name: typing.Optional[str] = None,
    wb_artifact_type_name: typing.Optional[str] = None,
    wb_entity_name: typing.Optional[str] = None,
    branch_name: typing.Optional[str] = None,
    metadata: typing.Optional[typing.Dict[str, typing.Any]] = None,
    assume_weave_type: typing.Optional[types.Type] = None,
    *,
    _lite_run: typing.Optional["InMemoryLazyLiteRun"] = None,
    _merge: typing.Optional[bool] = False,
) -> artifact_wandb.WandbArtifactRef:
    _orig_obj = obj
    _orig_ref = _get_ref(obj)
    if isinstance(_orig_ref, artifact_wandb.WandbArtifactRef):
        return _orig_ref
    if isinstance(_orig_ref, artifact_local.LocalArtifactRef):
        res = PUBLISH_CACHE_BY_LOCAL_ART.get(_orig_ref)
        if res is not None:
            ref_base._put_ref(obj, res)
            return res

    weave_type = assume_weave_type or _get_weave_type_with_refs(obj)

    target_project_from_context = get_publish_target_project()
    if target_project_from_context is not None:
        wb_project_name = target_project_from_context["project_name"]
    else:
        wb_project_name = wb_project_name or artifact_wandb.DEFAULT_WEAVE_OBJ_PROJECT
    name = name or _get_name(weave_type, obj)
    wb_artifact_type_name = wb_artifact_type_name or weave_type.root_type_class().name

    _assert_valid_artifact_name(name)
    _assert_valid_project_name(wb_project_name)
    _assert_valid_name_part(wb_artifact_type_name)
    _assert_valid_name_part(branch_name)
    _assert_valid_entity_name(wb_entity_name)
    # Validate entity name once we have them

    obj = box.box(obj)
    artifact = artifact_wandb.WandbArtifact(name, type=wb_artifact_type_name)
    artifact.metadata.update(metadata or {})

    # Use context to ensure any recursively published objects go to the same project
    with publish_target_project({"project_name": wb_project_name}):
        obj = _ensure_object_components_are_published(obj, weave_type, artifact)
    artifact_fs.update_weave_meta(weave_type, artifact)
    ref = artifact.set("obj", weave_type, obj)
    if not isinstance(ref, artifact_wandb.WandbArtifactRef):
        raise errors.WeaveSerializeError(
            "Expected a WandbArtifactRef, got %s" % type(ref)
        )

    # Only save if we have a ref into the artifact we created above. Otherwise
    # nothing new was created, so just return the existing ref.
    if ref.artifact == artifact:
        artifact.save(
            project=wb_project_name,
            entity_name=wb_entity_name,
            branch=branch_name,
            artifact_collection_exists=bool(_merge),
            _lite_run=_lite_run,
        )

    if isinstance(_orig_ref, artifact_local.LocalArtifactRef):
        PUBLISH_CACHE_BY_LOCAL_ART[_orig_ref] = ref

    ref_base._put_ref(obj, ref)
    ref_base._put_ref(_orig_obj, ref)

    return ref


def direct_save(
    obj: typing.Any,
    name: typing.Optional[str] = None,
    branch_name: typing.Optional[str] = None,
    source_branch_name: typing.Optional[str] = None,
    assume_weave_type: typing.Optional[types.Type] = None,
    artifact: typing.Optional[artifact_local.LocalArtifact] = None,
) -> artifact_local.LocalArtifactRef:
    weave_type = assume_weave_type or _get_weave_type_with_refs(obj)
    name = name or _get_name(weave_type, obj)

    _assert_valid_name_part(name)
    _assert_valid_name_part(branch_name)
    _assert_valid_name_part(source_branch_name)

    obj = box.box(obj)
    if artifact is None:
        # Using `version=source_branch_name` feels too magical. Would be better
        # to have a classmethod on LocalArtifact that takes a branch name and
        # returns an artifact with that branch name. This also precludes
        # directly saving an artifact with a branchpoint that is not local
        artifact = artifact_local.LocalArtifact(name, version=source_branch_name)
        artifact_fs.update_weave_meta(weave_type, artifact)
    ref = artifact.set("obj", weave_type, obj)

    # Only save if we have a ref into the artifact we created above. Otherwise
    #     nothing new was created, so just return the existing ref.
    if ref.artifact == artifact:
        artifact.save(branch=branch_name)  # type: ignore[no-untyped-call]

    return ref  # type: ignore


def publish(obj, name=None, type=None):  # type: ignore
    # We would probably refactor this method to be more like _direct_publish. This effectively
    # just a wrapper that let's the user specify project name with a slash.
    # TODO: should we only expose save for our API with a "remote" flag or something
    project_name = None
    if name is not None and "/" in name:
        project_name, name = name.split("/")

    return _direct_publish(
        obj,
        name=name,
        wb_project_name=project_name,
        assume_weave_type=type,
    )


def save(  # type: ignore
    obj,
    name=None,
    type=None,
    artifact=None,
    branch=None,
) -> artifact_local.LocalArtifactRef:
    # We would probably refactor this method to be more like direct_save. This effectively
    # just a wrapper that let's the user specify source_branch name with a slash.
    source_branch = None
    if name is not None and ":" in name:
        name, source_branch = name.split(":", 1)

    return direct_save(
        obj=obj,
        name=name,
        branch_name=branch,
        source_branch_name=source_branch,
        assume_weave_type=type,
        artifact=artifact,
    )


def get(uri_s: typing.Union[str, ref_base.Ref]) -> typing.Any:
    if isinstance(uri_s, ref_base.Ref):
        return uri_s.get()
    return ref_base.Ref.from_str(uri_s).get()


get_local_version_ref = artifact_local.get_local_version_ref
get_local_version = artifact_local.get_local_version


def deref(ref):  # type: ignore
    if isinstance(ref, ref_base.Ref):
        return ref.get()
    return ref


def _get_ref(obj: typing.Any) -> typing.Optional[ref_base.Ref]:
    if isinstance(obj, ref_base.Ref):
        return obj
    return ref_base.get_ref(obj)


def clear_ref(obj):  # type: ignore
    ref_base.clear_ref(obj)


get_ref = _get_ref


# Return all local artifacts.
# Warning: may iterate a lot of the filesystem!
def local_artifacts() -> typing.List[artifact_local.LocalArtifact]:
    result = []
    obj_paths = sorted(
        pathlib.Path(artifact_local.local_artifact_dir()).iterdir(),
        key=os.path.getctime,
    )
    for art_path in obj_paths:
        if os.path.basename(art_path) != "tmp":
            result.append(artifact_local.LocalArtifact(art_path.name, None))
    return result


def all_objects():  # type: ignore
    result = []
    obj_paths = sorted(
        pathlib.Path(artifact_local.local_artifact_dir()).iterdir(),
        key=os.path.getctime,
    )
    for art_path in obj_paths:
        ref = artifact_local.get_local_version_ref(art_path.name, "latest")
        if ref is not None:
            result.append((ref.created_at, ref))
    return [r[1] for r in sorted(result)]


def objects(
    of_type: types.Type, alias: str = "latest"
) -> typing.List[artifact_local.LocalArtifactRef]:
    result = []
    for art_name in os.listdir(artifact_local.local_artifact_dir()):
        try:
            ref = artifact_local.get_local_version_ref(art_name, alias)
            if ref is not None:
                if of_type.assign_type(ref.type):
                    # TODO: Why did I have this here?
                    # obj = ref.get()
                    # if isinstance(ref.type, types.RunType) and obj.op_name == "op-objects":
                    #     continue
                    result.append((ref.artifact.created_at, ref))
        except errors.WeaveSerializeError:
            # This happens because we may not have loaded ecosystem stuff that we need
            # to deserialize
            continue
    # Sorted by created_at
    return [r[1] for r in sorted(result)]


def recursively_unwrap_arrow(obj):  # type: ignore
    if getattr(obj, "to_pylist_notags", None):
        return obj.to_pylist_notags()
    if getattr(obj, "as_py", None):
        return obj.as_py()
    if isinstance(obj, graph.Node):
        return obj
    if isinstance(obj, dict):
        return {k: recursively_unwrap_arrow(v) for (k, v) in obj.items()}
    elif isinstance(obj, list):
        return [recursively_unwrap_arrow(item) for item in obj]
    return obj


def _default_ref_persister_artifact(
    type: types.Type, refs: typing.Iterable[artifact_base.ArtifactRef]
) -> artifact_base.Artifact:
    fs_art = artifact_local.LocalArtifact(type.name, "latest")
    # Save all the reffed objects into the new artifact.
    for mem_ref in refs:
        if mem_ref.path is not None and mem_ref._type is not None:
            fs_art.set(mem_ref.path, mem_ref._type, mem_ref._obj)
    fs_art.save()  # type: ignore[no-untyped-call]
    return fs_art


def to_python(
    obj: typing.Any,
    wb_type: typing.Optional[types.Type] = None,
    ref_persister: typing.Callable[
        [types.Type, typing.Iterable[artifact_base.ArtifactRef]], artifact_base.Artifact
    ] = _default_ref_persister_artifact,
) -> dict:
    if wb_type is None:
        wb_type = types.type_of_with_refs(obj)

    # First map the object using a MemArtifact to capture any custom object refs.
    art = artifact_mem.MemArtifact()
    mapper = mappers_python.map_to_python(wb_type, art)
    val = mapper.apply(obj)

    if art.ref_count() > 0:
        # There are custom objects, create a local artifact to persist them.
        fs_art = ref_persister(wb_type, art.refs())
        # now map the original object again. Because there are now existing refs
        # to the local artifact for any custom objects, this new value will contain
        # those existing refs as absolute refs. We provide None for artifact because
        # it should not be used in this pass.
        # TODO: actually had to add artifact here to get HF/Bertviz caching
        # to work. Why?
        mapper = mappers_python.map_to_python(wb_type, fs_art)  # type: ignore
        val = mapper.apply(obj)
    # TODO: this should be a ConstNode!
    return {"_type": wb_type.to_dict(), "_val": val}


def to_safe_const(obj):  # type: ignore
    wb_type = types.TypeRegistry.type_of(obj)
    mapper = mappers_python.map_to_python(wb_type, artifact_mem.MemArtifact())
    val = mapper.apply(obj)
    return graph.ConstNode(wb_type, val)


def from_python(obj: dict, wb_type: typing.Optional[types.Type] = None) -> typing.Any:
    if wb_type is None:
        wb_type = types.TypeRegistry.type_from_dict(obj["_type"])
    mapper = mappers_python.map_from_python(wb_type, None)  # type: ignore
    res = mapper.apply(obj["_val"])
    return res


def to_json_with_refs(  # type: ignore
    obj: typing.Any,
    artifact: artifact_base.Artifact,
    path: typing.Optional[list[str]] = None,
    wb_type: typing.Optional[types.Type] = None,
):
    """Convert an object to a JSON-serializable object, with refs to any custom objects.

    Contains business logic:
      - OpDef are always saved as top-level artifacts.
      - All other custom objects are saved as refs within the provided artifact
        if there is not are a ref available for them.
    """
    # This is newer than to_python, save and publish above, and doesn't use the "mapper"
    # pattern, which is overkill. Much better to just write a simple function like this.
    from weave.legacy.weave import op_def

    if wb_type is None:
        wb_type = types.TypeRegistry.type_of(obj)
    if path is None:
        path = []
    if isinstance(wb_type, (types.Number, types.String, types.NoneType, types.Boolean)):
        return obj
    elif isinstance(wb_type, types.UnknownType):
        raise errors.WeaveTypeError(f"Can't serialize {obj}. Type is UnknownType.")
    elif isinstance(wb_type, types.TypedDict):
        if not isinstance(obj, dict):
            raise ValueError("Expected dict for TypedDict, got %s" % type(obj).__name__)
        return {
            k: to_json_with_refs(
                v, artifact, path + [k], wb_type=wb_type.property_types[k]
            )
            for (k, v) in obj.items()
        }
    elif isinstance(wb_type, types.List):
        if not isinstance(obj, list):
            raise ValueError("Expected list for List, got %s" % type(obj).__name__)
        return [
            to_json_with_refs(v, artifact, path + [str(i)], wb_type=wb_type.object_type)
            for i, v in enumerate(obj)
        ]
    elif isinstance(obj, op_def.OpDef):
        try:
            gc = weave_client_context.require_weave_client()
        except errors.WeaveInitError:
            raise errors.WeaveSerializeError(
                "Can't serialize OpDef with a client initialization"
            )
        return gc._save_object(obj, obj.name, "latest")
    else:
        res = artifact.set("/".join(path), wb_type, obj)
        if res.artifact == artifact:
            res.serialize_as_path_ref = True
        return res


def make_js_serializer():  # type: ignore
    artifact = artifact_mem.MemArtifact()
    return functools.partial(to_weavejs, artifact=artifact)


def convert_timestamps_to_epoch_ms(obj: typing.Any) -> typing.Any:
    if isinstance(obj, list):
        return [convert_timestamps_to_epoch_ms(o) for o in obj]
    if isinstance(obj, dict):
        return {k: convert_timestamps_to_epoch_ms(v) for k, v in obj.items()}
    if isinstance(obj, datetime.datetime):
        return timestamp.python_datetime_to_ms(obj)
    return obj


def to_weavejs(obj, artifact: typing.Optional[artifact_base.Artifact] = None):  # type: ignore
    from weave.legacy.weave.arrow import list_ as arrow_list

    obj = box.unbox(obj)
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    elif isinstance(obj, dict):
        return {k: to_weavejs(v, artifact=artifact) for (k, v) in obj.items()}
    elif isinstance(obj, list):
        return [to_weavejs(item, artifact=artifact) for item in obj]
    elif isinstance(obj, arrow_list.ArrowWeaveList):
        return arrow_list.convert_arrow_timestamp_to_epoch_ms(obj).to_pylist_notags()

    wb_type = types.TypeRegistry.type_of(obj)

    if artifact is None:
        artifact = artifact_mem.MemArtifact()

    mapper = mappers_python.map_to_python(
        wb_type, artifact, mapper_options={"use_stable_refs": False}
    )
    return mapper.apply(obj)


def artifact_path_ref(artifact_path: str) -> artifact_base.ArtifactRef:
    """Return a reference to an object in the current artifact based on its path.

    This is only callable during artifact loading, and should not generally be
    called by user code.
    """
    loading_artifact = artifact_fs.get_loading_artifact()
    if loading_artifact is None:
        raise errors.WeaveSerializeError(
            "artifact_path_ref can only be used while loading an artifact."
        )

    return loading_artifact.ref_from_local_str(artifact_path)



import dataclasses
import typing

import weave
from weave.legacy.weave import panel

span_typed_dict_type = weave.types.TypedDict(
    {
        "name": weave.types.optional(weave.types.String()),
        "span_id": weave.types.optional(weave.types.String()),
        "parent_id": weave.types.optional(weave.types.String()),
        "trace_id": weave.types.optional(weave.types.String()),
        "start_time_s": weave.types.optional(weave.types.Number()),
        "end_time_s": weave.types.optional(weave.types.Number()),
        "status_code": weave.types.optional(weave.types.String()),
        "inputs": weave.types.optional(weave.types.TypedDict({})),
        "output": weave.types.optional(weave.types.Any()),
        "exception": weave.types.optional(weave.types.String()),
        "attributes": weave.types.optional(weave.types.TypedDict({})),
        "summary": weave.types.optional(weave.types.TypedDict({})),
        "timestamp": weave.types.optional(weave.types.Timestamp()),
    },
    not_required_keys=set(
        # parent_id is not required because it is None for root spans
        [
            "status_code",
            "inputs",
            "output",
            "exception",
            "attributes",
            "summary",
            "parent_id",
        ]
    ),
)


@weave.type("tracePanelConfig")
class PanelTraceConfig:
    selectedSpanIndex: typing.Optional[int] = 0


@weave.type("tracePanel")
class Trace(panel.Panel):
    id = "tracePanel"
    config: typing.Optional[PanelTraceConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = PanelTraceConfig()
        if "selectedSpanIndex" in options:
            self.config.selectedSpanIndex = options["selectedSpanIndex"]


@weave.op(
    name="panel_trace-active_span",
    output_type=weave.types.optional(span_typed_dict_type),
)
def active_span(self: Trace):
    index = 0 if self.config is None else self.config.selectedSpanIndex
    return weave.legacy.weave.ops_arrow.list_ops.index(self.input_node, index)



"""Common utilities for the LangChain integration.

This file exposes 4 primary functions:
- `safely_convert_lc_run_to_wb_span`: Converts a LangChain Run into a W&B Trace Span.
- `safely_get_span_producing_model`: Retrieves the model that produced a given LangChain Run.
- `safely_convert_model_to_dict`: Converts a LangChain model into a dictionary.

These functions are used by the `WandbTracer` to extract and save the relevant information.
"""

import logging
from typing import TYPE_CHECKING, Any, Optional, Union

from langchain.callbacks.tracers.schemas import Run, RunTypeEnum

from weave.legacy.weave.ops_domain import trace_tree

if TYPE_CHECKING:
    from langchain.chains.base import Chain
    from langchain.llms.base import BaseLLM
    from langchain.schema import BaseLanguageModel
    from langchain.tools.base import BaseTool

logger = logging.getLogger(__name__)

PRINT_WARNINGS = True


def safely_convert_lc_run_to_wb_span(run: Run) -> Optional["trace_tree.Span"]:
    try:
        return _convert_lc_run_to_wb_span(run)
    except Exception as e:
        if PRINT_WARNINGS:
            logging.warn(
                f"Skipping trace saving - unable to safely convert LangChain Run into W&B Trace due to: {e}"
            )
        return None


# def safely_get_span_producing_model(run: Run) -> Any:
#     try:
#         return run.serialized.get("_self")
#     except Exception as e:
#         if PRINT_WARNINGS:
#             wandb.termwarn(
#                 f"Skipping model saving - unable to safely retrieve LangChain model due to: {e}"
#             )
#     return None


# def safely_convert_model_to_dict(
#     model: Union["BaseLanguageModel", "BaseLLM", "BaseTool", "Chain"]
# ) -> Optional[dict]:
#     """Returns the model dict if possible, otherwise returns None.

#     Given that Models are all user defined, this operation is not always possible.
#     """
#     data = None
#     message = None
#     try:
#         data = model.dict()
#     except Exception as e:
#         message = str(e)
#         if hasattr(model, "agent"):
#             try:
#                 data = model.agent.dict()
#             except Exception as e:
#                 message = str(e)

#     if data is not None and not isinstance(data, dict):
#         message = (
#             f"Model's dict transformation resulted in {type(data)}, expected a dict."
#         )
#         data = None

#     if data is not None:
#         data = _replace_type_with_kind(data)
#     else:
#         if PRINT_WARNINGS:
#             wandb.termwarn(
#                 f"Skipping model saving - unable to safely convert LangChain Model to dictionary due to: {message}"
#             )

#     return data


def _convert_lc_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    if run.run_type == RunTypeEnum.llm:
        return _convert_llm_run_to_wb_span(run)
    elif run.run_type == RunTypeEnum.chain:
        return _convert_chain_run_to_wb_span(run)
    elif run.run_type == RunTypeEnum.tool:
        return _convert_tool_run_to_wb_span(run)
    else:
        return _convert_unknown_run_to_wb_span(run)


def _convert_llm_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _base_convert_run_to_wb_span(run)

    base_span["results"] = [
        trace_tree.Result(
            inputs={"prompt": prompt},
            outputs={
                f"gen_{g_i}": gen["text"]
                for g_i, gen in enumerate(run.outputs["generations"][ndx])
            }
            if (
                run.outputs is not None
                and len(run.outputs["generations"]) > ndx
                and len(run.outputs["generations"][ndx]) > 0
            )
            else None,
        )
        for ndx, prompt in enumerate(run.inputs["prompts"] or [])
    ]
    base_span["span_kind"] = trace_tree.SpanKind.LLM

    return trace_tree.Span(**base_span)


def _convert_chain_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _base_convert_run_to_wb_span(run)

    base_span["results"] = [trace_tree.Result(inputs=run.inputs, outputs=run.outputs)]
    base_span["child_spans"] = [
        _convert_lc_run_to_wb_span(child_run) for child_run in run.child_runs
    ]
    base_span["span_kind"] = (
        trace_tree.SpanKind.AGENT
        if "agent" in run.serialized.get("name").lower()
        else trace_tree.SpanKind.CHAIN
    )

    return trace_tree.Span(**base_span)


def _convert_tool_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _base_convert_run_to_wb_span(run)

    base_span["attributes"]["input"] = run.inputs["input"]
    base_span["results"] = [trace_tree.Result(inputs=run.inputs, outputs=run.outputs)]
    base_span["child_spans"] = [
        _convert_lc_run_to_wb_span(child_run) for child_run in run.child_runs
    ]
    base_span["span_kind"] = trace_tree.SpanKind.TOOL

    return trace_tree.Span(**base_span)


def _convert_unknown_run_to_wb_span(run: "Run") -> "trace_tree.Span":
    base_span = _base_convert_run_to_wb_span(run)

    return trace_tree.Span(**base_span)


def _base_convert_run_to_wb_span(run: "Run") -> dict:
    attributes = {**run.extra} if run.extra else {}
    attributes["execution_order"] = run.execution_order

    return dict(
        span_id=str(run.id) if run.id is not None else None,
        _name=run.serialized.get("name"),
        start_time_ms=int(run.start_time.timestamp() * 1000),
        end_time_ms=int(run.end_time.timestamp() * 1000),
        status_code=trace_tree.StatusCode.SUCCESS
        if run.error is None
        else trace_tree.StatusCode.ERROR,
        status_message=run.error,
        attributes=attributes,
    )


def _replace_type_with_kind(data: dict) -> dict:
    if isinstance(data, dict):
        # W&B TraceTree expects "_kind" instead of "_type" since `_type` is special
        # in W&B.
        if "_type" in data:
            _type = data.pop("_type")
            data["_kind"] = _type
        return {k: _replace_type_with_kind(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_replace_type_with_kind(v) for v in data]
    elif isinstance(data, tuple):
        return tuple(_replace_type_with_kind(v) for v in data)
    elif isinstance(data, set):
        return {_replace_type_with_kind(v) for v in data}
    else:
        return data



import pytest

import weave
from weave.legacy.weave import ops_arrow

LIST_CONSTRUCTORS = [
    lambda x: x,
    lambda x: ops_arrow.to_arrow(x),
]


@pytest.mark.parametrize("self_constructor", LIST_CONSTRUCTORS)
def test_index_simple(self_constructor):
    items = weave.save(self_constructor(["a", "b", "c"]))
    assert weave.use(items[0]) == "a"


@pytest.mark.parametrize("self_constructor", LIST_CONSTRUCTORS)
def test_index_negative(self_constructor):
    items = weave.save(self_constructor(["a", "b", "c"]))
    assert weave.use(items[-1]) == "c"


@pytest.mark.parametrize("self_constructor", LIST_CONSTRUCTORS)
def test_index_list(self_constructor):
    items = weave.save(self_constructor(["a", "b", "c"]))
    res = weave.use(items[[0, 2]])
    if isinstance(res, ops_arrow.ArrowWeaveList):
        res = res.to_pylist_notags()
    assert res == ["a", "c"]


@pytest.mark.parametrize("self_constructor", LIST_CONSTRUCTORS)
def test_index_arrowlist(self_constructor):
    items = weave.save(self_constructor(["a", "b", "c"]))
    res = weave.use(items[ops_arrow.to_arrow([0, 2])])
    if isinstance(res, ops_arrow.ArrowWeaveList):
        res = res.to_pylist_notags()
    assert res == ["a", "c"]


def test_index_listawl_oob():
    val = ops_arrow.to_arrow([[0, 1], [2, 3]])
    val.object_type = ops_arrow.ArrowWeaveListType(val.object_type.object_type)
    items = weave.save(val)
    res = weave.use(items[2])
    if isinstance(res, ops_arrow.ArrowWeaveList):
        res = res.to_pylist_notags()
    assert res == None



import weave
from weave.legacy.weave import panel


@weave.type()
class PanelNumber(panel.Panel):
    id = "number"


@weave.type()
class PanelBoolean(panel.Panel):
    id = "boolean"


@weave.type()
class PanelDate(panel.Panel):
    id = "date"
